import munch

try:
    import sdds
except ImportError:
    print("No SDDS available!")

import numpy as np
import enum


def read_sdds_file(filename, ascii=False, object=None):
    sddsobject = SDDSFile(index=0, ascii=ascii)
    sddsobject.read_file(filename)
    if object is None:
        object = munch.Munch()
        for k, v in sddsobject.data.items():
            object[k] = v
        return object
    return sddsobject


class MyEnumMeta(enum.EnumMeta):
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        else:
            return True


class SDDS_Types(enum.IntEnum, metaclass=MyEnumMeta):
    SDDS_LONGDOUBLE = 1
    SDDS_DOUBLE = 2
    SDDS_REAL64 = 2
    SDDS_FLOAT = 3
    SDDS_REAL32 = 3
    SDDS_LONG = 4
    SDDS_INT32 = 4
    SDDS_ULONG = 5
    SDDS_UINT32 = 5
    SDDS_SHORT = 6
    SDDS_INT16 = 6
    SDDS_USHORT = 7
    SDDS_UINT16 = 7
    SDDS_STRING = 8
    SDDS_CHARACTER = 9
    SDDS_NUM_TYPES = 9
    SDDS_BINARY = 1
    SDDS_ASCII = 2


class SDDSObject(munch.Munch):

    def __init__(
        self,
        index=1,
        name=None,
        data=[],
        unit="",
        type=2,
        symbol="",
        formatstring="",
        fieldlength=0,
        description="",
    ):
        super().__init__()
        self._types = SDDS_Types
        self._name = name
        self._data = data
        self._unit = unit
        self._symbol = symbol
        self._type = type
        self._formatstring = formatstring
        self._fieldlength = fieldlength
        self._description = description

    def __repr__(self):
        return repr(
            {
                "name": self.name,
                "unit": self.unit,
                "type": self._types(self.type).name,
                "data": self.data,
            }
        )

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name
        return self._name

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit):
        if unit is not None:
            self._unit = unit
        return self._unit

    @property
    def symbol(self):
        return self._symbol

    @symbol.setter
    def symbol(self, symbol):
        if symbol is not None:
            self._symbol = symbol
        return self._symbol

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        if isinstance(type, str):
            if type in self._types:
                self._type = type
        elif isinstance(type, int):
            self._type = type
        return self._type

    @property
    def fieldlength(self):
        return self._fieldlength

    @fieldlength.setter
    def fieldlength(self, length):
        if isinstance(length, (int, float)):
            self._fieldlength = int(fieldlength)
        return self._fieldlength

    @property
    def formatstring(self):
        return self._formatstring

    @formatstring.setter
    def formatstring(self, string):
        if isinstance(string, str):
            self._formatstring = string
        return self._formatstring

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, string):
        if isinstance(string, str):
            self._description = string
        return self._description

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data
        return self._data


class SDDSColumn(SDDSObject):

    def __init__(
        self,
        name=None,
        data=[],
        unit="",
        type=2,
        symbol="",
        formatstring="",
        fieldlength=0,
        description="",
    ):
        super().__init__(name=name, data=None, unit=unit, type=type, symbol=symbol)
        self.objectType = "Column"
        self.data = data

    @property
    def data(self):
        if isinstance(self._data, (np.ndarray)):
            return self._data.tolist()
        return self._data

    @data.setter
    def data(self, data):
        if not isinstance(data, (dict, tuple, list, np.ndarray)):
            if isinstance(data, (float, int, str)):
                data = [data]
            else:
                raise Exception("Wrong data type for SDDS Column!", type(data))
        self._data = data
        return self._data

    def length(self):
        return len(self._data)


class SDDSParameter(SDDSObject):

    def __init__(
        self,
        name=None,
        data=[],
        unit="",
        type=2,
        symbol="",
        formatstring="",
        fieldlength=0,
        description="",
    ):
        super().__init__(name=name, data=None, unit=unit, type=type, symbol=symbol)
        self.objectType = "Parameter"
        self.data = data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if not isinstance(data, (dict, tuple, list)):
            self._data = [data]
        else:
            raise Exception("Wrong data type for SDDS Parameter!", type(data))
        return self._data

    @property
    def fieldlength(self):
        if self._fieldlength != 0:
            return self._fieldlength
        else:
            return ""


class SDDSArray(munch.Munch):

    def __init__(self, columns):
        super().__init__(columns)

    def __getitem__(self, itemkey):
        try:
            return {
                k: getattr(v, itemkey) if hasattr(v, itemkey) else v[itemkey]
                for k, v in self.items()
            }
        except:
            try:
                return super().__getitem__(itemkey)
            except KeyError:
                raise AttributeError(itemkey)


class SDDSFile(object):

    def __init__(self, index=1, ascii=False):
        super().__init__()
        self._types = SDDS_Types
        self._columns = munch.Munch()
        self._parameters = munch.Munch()
        self._index = index
        try:
            self._sddsObject = sdds.SDDS(0)
        except:
            self._sddsObject = sdds.sdds.SDDS()
        if ascii:
            self._sddsObject.mode = self._sddsObject.SDDS_ASCII
        else:
            self._sddsObject.mode = self._sddsObject.SDDS_BINARY

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        self._index = index
        return self._index

    def clear(self):
        self._columns = munch.Munch()
        self._parameters = munch.Munch()
        try:
            self._sddsObject = sdds.SDDS(None)
        except:
            self._sddsObject = sdds.sdds.SDDS(None)

    def column_names(self):
        return self._columns.keys()

    def parameter_names(self):
        return self._parameters.keys()

    def columns(self):
        return SDDSArray(self._columns)

    def parameters(self):
        return SDDSArray(self._parameters)

    def add_column(
        self,
        name,
        data,
        type=2,
        unit="",
        symbol="",
        formatstring="",
        fieldlength=0,
        description="",
    ):
        self._columns[name] = SDDSColumn(
            name=name,
            data=data,
            unit=unit,
            type=type,
            symbol=symbol,
            formatstring=formatstring,
            fieldlength=fieldlength,
            description=description,
        )

    def add_columns(
        self,
        name,
        data,
        type,
        unit,
        symbol,
        formatstring=None,
        fieldlength=None,
        description=None,
    ):
        combined_data = {
            k: v
            for k, v in {
                "name": name,
                "data": data,
                "type": type,
                "unit": unit,
                "symbol": symbol,
                "formatstring": formatstring,
                "fieldlength": fieldlength,
                "description": description,
            }.items()
            if v is not None
        }
        for i in range(len(name)):
            data = dict(
                zip(
                    combined_data.keys(),
                    [combined_data[k][i] for k in combined_data.keys()],
                )
            )
            self.add_column(**data)

    def add_parameter(
        self,
        name,
        data,
        type=2,
        unit="",
        symbol="",
        formatstring="",
        fieldlength=0,
        description="",
    ):
        self._parameters[name] = SDDSParameter(
            name=name,
            data=data,
            unit=unit,
            type=type,
            symbol=symbol,
            formatstring=formatstring,
            fieldlength=fieldlength,
            description=description,
        )

    def add_parameters(
        self,
        name,
        data,
        type,
        unit,
        symbol,
        formatstring=None,
        fieldlength=None,
        description=None,
    ):
        combined_data = {
            k: v
            for k, v in {
                "name": name,
                "data": data,
                "type": type,
                "unit": unit,
                "symbol": symbol,
                "formatstring": formatstring,
                "fieldlength": fieldlength,
                "description": description,
            }.items()
            if v is not None
        }
        for i in range(len(name)):
            data = dict(
                zip(
                    combined_data.keys(),
                    [combined_data[k][i] for k in combined_data.keys()],
                )
            )
            self.add_parameter(**data)

    def save(self, *args, **kwargs):
        return self.write_file(*args, **kwargs)

    def dimensions(self, list):
        l = list
        for i in range(5):
            try:
                l = l[0]
            except:
                return i

    def write_file(self, filename):
        for name, param in self._parameters.items():
            self._sddsObject.defineParameter(
                param.name,
                param.symbol,
                param.unit,
                param.description,
                param.formatstring,
                param.type,
                param.fieldlength,
            )
            self._sddsObject.setParameterValueList(param.name, param.data)
        for name, column in self._columns.items():
            # print(len([list(column.data)][0]))
            self._sddsObject.defineColumn(
                column.name,
                column.symbol,
                column.unit,
                column.description,
                column.formatstring,
                column.type,
                column.fieldlength,
            )
            self._sddsObject.setColumnValueLists(column.name, [column.data])
        self._sddsObject.save(filename)

    def load(self, *args, **kwargs):
        return self.read_file(*args, **kwargs)

    def read_file(self, filename, page=-1):
        self._sddsObject.load(filename)
        sddsref = self._sddsObject
        for col in range(len(sddsref.columnName)):
            symbol, unit, description, formatString, type, fieldLength = (
                sddsref.columnDefinition[col]
            )
            column_data = np.array(sddsref.columnData[col][page])
            self.add_column(
                sddsref.columnName[col],
                column_data,
                type=type,
                unit=unit,
                symbol=symbol,
                formatstring=formatString,
                fieldlength=fieldLength,
                description=description,
            )
        # sddsobject.SDDSparameterNames = list()
        for param in range(len(sddsref.parameterName)):
            name = sddsref.parameterName[param]
            symbol, unit, description, formatString, type, fieldLength = (
                sddsref.parameterDefinition[param]
            )
            parameter_data = sddsref.parameterData[param]
            self.add_parameter(
                sddsref.parameterName[param],
                parameter_data[page],
                type=type,
                unit=unit,
                symbol=symbol,
                formatstring=formatString,
                fieldlength=fieldLength,
                description=description,
            )

    @property
    def data(self):
        return {k: v.data for k, v in {**self._parameters, **self._columns}.items()}



import sys
from pprint import pprint

sys.path.append("..")
from PAdantic.PAdantic import PAdantic  # noqa E402

machine = PAdantic(
    layout="../../padantic-lattices/CLARA/layouts.yaml",
    section="../../padantic-lattices/CLARA/sections.yaml",
    element_list="../../padantic-lattices/NEW/CLARA/YAML/",
)


def flatten(xss):
    return [x for xs in xss for x in xs]


def unique_list(a_list):
    seen = list()
    for x in flatten(a_list):
        if x not in seen:
            seen.append(x)
    return seen


machine.default_path = "SP3"
print(machine.get_elements_s_pos())
pprint({d: machine.get_element(d).base_model_dump() for d in machine.get_dipoles()})
# exit()
for magnet_name in machine.get_quadrupoles():
    linsat = machine.get_element(magnet_name).magnetic.linear_saturation_coefficients
    kvals = magnet_name, linsat.currentToK(48, 30)
    print(magnet_name, kvals[1]["K"], linsat.KToCurrent(kvals[1], 30))
    print(magnet_name, kvals[1]["KL"], linsat.KLToCurrent(kvals[1], 30))
exit()



import os
from copy import copy
from typing import Any, Union
import yaml
from IPython.core.magic import Bunch
from pydantic import BaseModel, field_validator, ValidationInfo, Field

from .Magnet_Table import add_magnet_table_parameters
from ..models.PV import (  # noqa
    MagnetPV,
    BPMPV,
    BAMPV,
    BLMPV,
    CameraPV,
    ScreenPV,
    ChargeDiagnosticPV,
    VacuumGaugePV,
    LaserEnergyMeterPV,
    LaserHWPPV,
    LaserMirrorPV,
    LightingPV,
    PIDPV,
    LLRFPV,
    RFModulatorPV,
    ShutterPV,
    ValvePV,
    RFProtectionPV,
    RFHeartbeatPV,
    PV,
    elementTypes,
    PVTypes,
)
from ..models.element import (
    Quadrupole,
    Dipole,
    Sextupole,
    Solenoid,
    Marker,
    Aperture,
    Collimator,
    Beam_Position_Monitor,
    Beam_Arrival_Monitor,
    Bunch_Length_Monitor,
    Wall_Current_Monitor,
    Faraday_Cup_Monitor,
    Integrated_Current_Transformer,
    Screen,
    Combined_Corrector,
    Horizontal_Corrector,
    Vertical_Corrector,
    Wakefield,
    RFCavity,
    RFDeflectingCavity,
    Shutter,
    Camera,
)
from ..models.diagnostic import Camera_Diagnostic_Type

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/camera_assignments.yaml", "r"
) as stream:
    camera_assignments = yaml.load(stream, Loader=yaml.Loader)
    camera_types = {}
    for k, v in camera_assignments.items():
        for s in v:
            camera_types[s] = k


class SimFrame_Conversion(BaseModel):
    typeclass: Any
    PV_class: str = None
    hardware_class: str
    hardware_type: Union[str, None] = Field(validate_default=True, default=None)

    @field_validator("hardware_type", mode="before")
    def check_type(cls, value: Any, info: ValidationInfo) -> str:
        if value is None:
            value = info.data["typeclass"].model_fields["hardware_type"].default
        return value or None


SimFrame_Elements = {
    "quadrupole": SimFrame_Conversion(
        typeclass=Quadrupole, hardware_class="Magnet"
    ),
    "dipole": SimFrame_Conversion(
        typeclass=Dipole, hardware_class="Magnet"
    ),
    "sextupole": SimFrame_Conversion(
        typeclass=Sextupole, hardware_class="Magnet"
    ),
    "solenoid": SimFrame_Conversion(
        typeclass=Solenoid, hardware_class="Magnet"
    ),
    "marker": SimFrame_Conversion(
        typeclass=Marker, hardware_class="Simulation"
    ),
    "aperture": SimFrame_Conversion(
        typeclass=Aperture, hardware_class="Simulation"
    ),
    "collimator": SimFrame_Conversion(
        typeclass=Collimator, hardware_class="Simulation"
    ),
    "beam_position_monitor": SimFrame_Conversion(
        typeclass=Beam_Position_Monitor, hardware_class="Diagnostic"
    ),
    "beam_arrival_monitor": SimFrame_Conversion(
        typeclass=Beam_Arrival_Monitor, hardware_class="Diagnostic"
    ),
    "bunch_length_monitor": SimFrame_Conversion(
        typeclass=Bunch_Length_Monitor, hardware_class="Diagnostic"
    ),
    "wall_current_monitor": SimFrame_Conversion(
        typeclass=Wall_Current_Monitor, hardware_class="Diagnostic"
    ),
    "faraday_cup": SimFrame_Conversion(
        typeclass=Faraday_Cup_Monitor, hardware_class="Diagnostic"
    ),
    "integrated_current_transformer": SimFrame_Conversion(
        typeclass=Integrated_Current_Transformer, hardware_class="Diagnostic"
    ),
    "screen": SimFrame_Conversion(
        typeclass=Screen, hardware_class="Diagnostic"
    ),
    # 'rf_deflecting_cavity': SimFrame_Conversion(typeclass=Sextupole, PV_class='Magnet'),
    "kicker": SimFrame_Conversion(
        typeclass=Combined_Corrector, hardware_class="Magnet"
    ),
    "hkicker": SimFrame_Conversion(
        typeclass=Horizontal_Corrector, hardware_class="Magnet"
    ),
    "vkicker": SimFrame_Conversion(
        typeclass=Vertical_Corrector, hardware_class="Magnet"
    ),
    # 'monitor': SimFrame_Conversion(typeclass=Sextupole, PV_class='Magnet'),
    "longitudinal_wakefield": SimFrame_Conversion(
        typeclass=Wakefield, hardware_class="Simulation"
    ),
    "cavity": SimFrame_Conversion(
        typeclass=RFCavity, hardware_class="RF"
    ),
    "rf_deflecting_cavity": SimFrame_Conversion(
        typeclass=RFDeflectingCavity, hardware_class="RF"
    ),
    "shutter": SimFrame_Conversion(
        typeclass=Shutter, hardware_class="Vacuum"
    ),
}


def get_SimFrame_YAML_filename(original, replacement):
    splitstr = original.replace("\\", "/").split("/")
    idx = splitstr.index("YAML")
    return "/".join(splitstr[:idx]) + "/" + replacement


def get_SimFrame_MachineArea(name):
    return name.split("-")[1]


def get_SimFrame_PV(name):
    return name


def interpret_SimFrame_Element(name, elem):
    if "type" in elem and elem["type"] in SimFrame_Elements:
        # try:
        # print('type',elem['type'],'found')
        felem = SimFrame_Elements[elem["type"]].typeclass
        hasPV = SimFrame_Elements[elem["type"]].PV_class in PVTypes

        elem.update(dict(SimFrame_Elements[elem["type"]]))
        elem.update(
            {
                "name": name,
                "machine_area": get_SimFrame_MachineArea(name),
                "hardware_class": SimFrame_Elements[elem["type"]].hardware_class,
            }
        )

        fields = elem
        if hasPV:
            try:
                fpv = globals()[PVTypes[SimFrame_Elements[elem["type"]].PV_class]]
                elemPV = fpv.with_defaults(name)
                fields["controls"] = elemPV
            except Exception as e:
                print("interpret_SimFrame_Element", name, elem, fpv)
                raise e
        fields.update(
            **{
                k: v.annotation.from_CATAP(elem)
                for k, v in felem.model_fields.items()
                if hasattr(v.annotation, "from_CATAP")
            }
        )
        elemmodel = felem(**fields)
        if SimFrame_Elements[elem["type"]].hardware_class == "Magnet":
            elemmodel = add_magnet_table_parameters(
                name, elemmodel, get_SimFrame_PV(name)
            )
        return elemmodel
    # except Exception as e:
    #     print('Error', name, e)


def read_SimFrame_YAML(filename):
    # print('File:',filename)
    elemlist = {}
    with open(filename, "r") as stream:
        data = yaml.load(stream, Loader=yaml.Loader)
    for name, elem in data["elements"].items():
        if name == "filename":  # and isinstance(elem, str):
            if isinstance(elem, str):
                newfilename = get_SimFrame_YAML_filename(filename, elem)
                elemlist.update(read_SimFrame_YAML(newfilename))
            elif isinstance(elem, list):
                for e in elem:
                    newfilename = get_SimFrame_YAML_filename(filename, e)
                    elemlist.update(read_SimFrame_YAML(newfilename))
        elif "type" in elem and elem["type"] in SimFrame_Elements:
            if elem["type"] == "kicker":
                helem = copy(elem)
                helem["type"] = "hkicker"
                helem["mag_type"] = "HORIZONTAL_CORRECTOR"
                hname = name.replace("HVCOR", "HCOR")
                velem = copy(elem)
                velem["type"] = "vkicker"
                velem["mag_type"] = "VERTICAL_CORRECTOR"
                vname = name.replace("HVCOR", "VCOR")
                elemmodel = interpret_SimFrame_Element(hname, helem)
                elemlist.update({hname: elemmodel})
                elemmodel = interpret_SimFrame_Element(vname, velem)
                elemlist.update({vname: elemmodel})
                elem["Horizontal_Corrector"] = hname
                elem["Vertical_Corrector"] = vname
            if elem["type"] == "screen":
                elemmodel = interpret_SimFrame_Element(name, elem)
                elemlist.update({name: elemmodel})
                camtype = (
                    camera_types[elemmodel.name]
                    if elemmodel.name in camera_types
                    else "PCO"
                )
                print(elemmodel.diagnostic.camera_name, "camtype = ", camtype)
                elemmodelcam = Camera(
                    name=elemmodel.diagnostic.camera_name,
                    hardware_model=camtype,
                    machine_area=elemmodel.machine_area,
                    physical=elemmodel.physical,
                    diagnostic=Camera_Diagnostic_Type(type=camtype),
                    controls=None,
                )
                elemlist.update({elemmodel.diagnostic.camera_name: elemmodelcam})
            else:
                elemmodel = interpret_SimFrame_Element(name, elem)
                elemlist.update({name: elemmodel})
        else:
            # pass
            print("read_SimFrame_YAML", name, elem["type"])
        if "sub_elements" in elem:
            for subname, subelem in elem["sub_elements"].items():
                if "type" in subelem and subelem["type"] in SimFrame_Elements:
                    # print('Subelement:', subelem)
                    subelem["subelement"] = True
                    elemmodel = interpret_SimFrame_Element(subname, subelem)
                    # print(subname, elemmodel)
                    elemlist.update({subname: elemmodel})
    # print('simframe',elemlist)
    return elemlist


SF_files = [
    r"../../masterlattice/MasterLattice/YAML/CLA_Gun400.yaml",
    r"../../masterlattice/MasterLattice/YAML/CLA_SP2.yaml",
    r"../../masterlattice/MasterLattice/YAML/CLA_SP3.yaml",
    r"../../masterlattice/MasterLattice/YAML/CLA_FEBE.yaml",
    r"../../masterlattice/MasterLattice/YAML/CLA_SP1.yaml",
]



import builtins
from pydantic import (
    BaseModel,
    ValidationError,
    field_validator,
    model_serializer,
    ConfigDict,
)
from typing import Dict, Type


class ControlVariable(BaseModel):
    """
    Model representing a control variable in a system.
    """
    identifier: str
    """Unique identifier for the control variable."""

    dtype: type
    """Data type of the control variable (e.g., int, float, str)."""

    protocol: str
    """Protocol or method used to interact with the control variable."""

    units: str = "Arb. Units"
    """Units of measurement for the control variable."""

    description: str = "Default Description"
    """Description of the control variable."""

    read_only: bool = True
    """Indicates if the variable is read-only."""

    model_config = ConfigDict(
        arbitrary_types_allowed=False,
        extra="allow",
        frozen=True,
    )

    def __init__(self, **data):
        super().__init__(**data)

    @field_validator("dtype", mode="before")
    def validate_dtype(cls, v) -> Type:
        """Convert from string to type if necessary."""
        if isinstance(v, str):
            try:
                return getattr(builtins, v)
            except AttributeError:
                raise ValueError(f"Unknown dtype string: {v}")
        if isinstance(v, type):
            return v
        raise TypeError(f"dtype must be a type or string, got {type(v)}")

    @model_serializer
    def serialize(self):
        data = self.__dict__.copy()
        if isinstance(self.dtype, type):
            data["dtype"] = self.dtype.__name__
        elif isinstance(self.dtype, str):
            data["dtype"] = self.dtype
        return data


class ControlsInformation(BaseModel):
    """
    Model representing a collection of control variables.
    """

    variables: Dict[str, ControlVariable]
    """Dictionary mapping variable names to `~nala.models.control.ControlVariable` instances."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="allow",
        frozen=True,
    )

    @field_validator("variables", mode="before")
    def validate_variables(cls, v) -> Dict[str, ControlVariable]:
        """Ensure all values are ControlVariable instances."""
        if isinstance(v, dict):
            validated_dict = {}
            for key, value in v.items():
                if isinstance(value, ControlVariable):
                    validated_dict[key] = value
                elif isinstance(value, dict):
                    try:
                        validated_dict[key] = ControlVariable(**value)
                    except ValidationError as e:
                        raise ValueError(
                            (
                                "Invalid ControlVariable definition for key "
                                + f"'{key}': "
                                + f"{e}"
                            )
                        ) from e
                else:
                    raise TypeError(
                        (
                            f"Value for key '{key}'"
                            + "must be a ControlVariable or dict, "
                            + f"got {type(value)}"
                        )
                    )
            return validated_dict
        raise TypeError(f"variables must be a dict, got {type(v)}")



import os
import yaml
from typing import Union
from ..models.elementList import MachineModel
from ..models.element import PhysicalElement

def represent_tuple(dumper, data):
    return dumper.represent_sequence('tag:yaml.org,2002:seq', data)

yaml.add_representer(tuple, represent_tuple)


def export_as_yaml(
    filename: Union[str, None], ele: PhysicalElement = PhysicalElement
) -> None:
    if filename is not None:
        with open(filename, "w") as yaml_file:
            yaml.default_flow_style = False
            dump = ele.base_model_dump()
            # dump["hardware_subclass"] = ele.__class__.__name__
            dump.pop("CASCADING_RULES")
            yaml.dump(dump, yaml_file)
    else:
        dump = ele.base_model_dump()
        # dump["hardware_subclass"] = ele.__class__.__name__
        return dump


def export_machine_combined_file(path: str, machine: MachineModel) -> None:
    filename = os.path.join(path, "summary.yaml")
    os.makedirs(path, exist_ok=True)
    combined_yaml = {}
    for name, elem in machine.elements.items():
        combined_yaml[name] = export_as_yaml(None, elem)
    with open(filename, "w") as yaml_file:
        yaml.default_flow_style = True
        yaml.dump(combined_yaml, yaml_file)


def export_machine(path: str, machine: MachineModel, overwrite: bool = False, verbose: bool = False) -> None:
    os.makedirs(path, exist_ok=True)
    for name, elem in machine.elements.items():
        directory = os.path.join(path, elem.subdirectory)
        os.makedirs(directory, exist_ok=True)
        filename = os.path.join(directory, elem.name + ".yaml")
        if overwrite or not os.path.isfile(filename):
            if verbose:
                print("Exporting Element", name, "to file", filename)
            export_as_yaml(filename, elem)


def export_elements(path: str, elements: list[PhysicalElement]) -> None:
    for elem in elements:
        directory = os.path.join(path, elem.subdirectory)
        os.makedirs(directory, exist_ok=True)
        filename = os.path.join(directory, elem.name + ".yaml")
        export_as_yaml(filename, elem)



from typing import Dict, Any
from nala.models.elementList import MachineModel
from .converter import translate_elements
from .layout import MachineLayoutTranslator

class MachineModelTranslator(MachineModel):
    directory: str = '.'

    @classmethod
    def from_machine(cls, machine: MachineModel) -> "MachineModelTranslator":
        return cls.model_validate(
            {
                "layout": machine.model_copy().layout,
                "section": machine.model_copy().section,
                "elements": machine.model_copy().elements,
                "sections": machine.model_copy().sections,
                "lattices": machine.model_copy().lattices,
                "master_lattice_location": machine.model_copy().master_lattice_location,
            }
        )

    def to_astra(self) -> Dict[str, Dict[str, str]]:
        model = {}
        for name, latt in self.lattices.items():
            model.update(
                {
                    name: MachineLayoutTranslator.from_layout(latt).to_astra()
                }
            )
        return model

    def to_elegant(self, string: str = "", charge: float = None) -> str:
        for latt in self.lattices.values():
            for section in latt.sections.values():
                section_with_drifts = section.createDrifts()
                elem_dict = translate_elements(
                    section_with_drifts.values(),
                    master_lattice_location=self.master_lattice_location,
                    directory=self.directory,
                )
                if charge:
                    string += f"{section.name}_Q: CHARGE, TOTAL = {charge};\n"

                for d in elem_dict.values():
                    string += d.to_elegant()

                string += f"\n{section.name}: LINE = ("
                if charge:
                    string += f"{section.name}_Q, "
                for elem in section_with_drifts.keys():
                    string += f"{elem}, "
                string = f"{string[:-2]})" + "\n\n\n"

        for name, latt in self.lattices.items():
            string += f"{name}: LINE = ("
            for l in list(latt.keys()):
                string += f"{l}, "
            string = f"{string[:-2]})" + "\n\n"
        return string

    def to_genesis(self, string: str = "") -> str:
        for latt in self.lattices.values():
            for section in latt.sections.values():
                section_with_drifts = section.createDrifts()
                elem_dict = translate_elements(
                    section_with_drifts.values(),
                    master_lattice_location=self.master_lattice_location,
                    directory=self.directory,
                )

                for d in elem_dict.values():
                    string += d.to_genesis()

                string += f"\n{section.name}: LINE = " + "{"
                for elem in section_with_drifts.keys():
                    string += f"{elem}, "
                string = f"{string[:-2]}" + "}\n\n\n"

        for name, latt in self.lattices.items():
            string += f"{name}: LINE = " + "{"
            for l in list(latt.keys()):
                string += f"{l}, "
            string = f"{string[:-2]}" + "};\n\n"
        return string

    def to_ocelot(self, save=False) -> Dict[str, Dict[str, "MagneticLattice"]]:
        model = {}
        for name, latt in self.lattices.items():
            model.update(
                {
                    name: MachineLayoutTranslator.from_layout(latt).to_ocelot(save=save)
                }
            )
        return model

    def to_cheetah(self, save=False) -> Dict[str, Dict[str, "Segment"]]:
        model = {}
        for name, latt in self.lattices.items():
            model.update(
                {
                    name: MachineLayoutTranslator.from_layout(latt).to_cheetah(save=save)
                }
            )
        return model

    def to_xsuite(
            self,
            beam_length: int,
            env: Any = None,
            particle_ref: Any = None,
            save=False
    ) -> Dict[str, Dict[str, object]]:
        model = {}
        for name, latt in self.lattices.items():
            model.update(
                {
                    name: MachineLayoutTranslator.from_layout(latt).to_xsuite(
                        beam_length=beam_length,
                        env=env,
                        particle_ref=particle_ref,
                        save=save,
                    )
                }
            )
        return model


from typing import Dict, Any
from nala.models.elementList import MachineLayout
from .converter import translate_elements
from .section import SectionLatticeTranslator

class MachineLayoutTranslator(MachineLayout):
    directory: str = '.'

    @classmethod
    def from_layout(cls, layout: MachineLayout) -> "MachineLayoutTranslator":
        return cls.model_validate(
            {
                "name": layout.model_copy().name,
                "sections": layout.model_copy().sections,
                "master_lattice_location": layout.model_copy().master_lattice_location,
            }
        )

    def to_astra(self) -> Dict[str, str]:
        lattices = {}
        for section in self.sections.values():
            lattices.update(
                {
                    section.name: SectionLatticeTranslator.from_section(section).to_astra()
                }
            )
        return lattices

    def to_elegant(self, string: str = "", charge: float = None) -> str:
        for section in self.sections.values():
            section_with_drifts = section.createDrifts()
            elem_dict = translate_elements(
                section_with_drifts.values(),
                master_lattice_location=self.master_lattice_location,
                directory=self.directory,
            )
            if charge:
                string += f"{section.name}_Q: CHARGE, TOTAL = {charge};\n"

            for d in elem_dict.values():
                string += d.to_elegant()

            string += f"\n{section.name}: LINE = ("
            if charge:
                string += f"{section.name}_Q, "
            for elem in section_with_drifts.keys():
                string += f"{elem}, "
            string = f"{string[:-2]})" + "\n\n\n"
        return string

    def to_genesis(self, string: str = "") -> str:
        for section in self.sections.values():
            section_with_drifts = section.createDrifts()
            elem_dict = translate_elements(
                section_with_drifts.values(),
                master_lattice_location=self.master_lattice_location,
                directory=self.directory,
            )

            for d in elem_dict.values():
                string += d.to_genesis()

            string += f"\n{section.name}: LINE = " + "{"
            for elem in section_with_drifts.keys():
                string += f"{elem}, "
            string = f"{string[:-2]}" + "};\n\n\n"
        return string

    def to_ocelot(self, save=False) -> Dict[str, "MagneticLattice"]:
        lattices = {}
        for section in self.sections.values():
            lattices.update(
                {
                    section.name: SectionLatticeTranslator.from_section(section).to_ocelot(save=save)
                }
            )
        return lattices

    def to_cheetah(self, save=False) -> Dict[str, "Segment"]:
        lattices = {}
        for section in self.sections.values():
            lattices.update(
                {
                    section.name: SectionLatticeTranslator.from_section(section).to_cheetah(save=save)
                }
            )
        return lattices

    def to_xsuite(
            self,
            beam_length: int,
            env: Any = None,
            particle_ref: Any = None,
            save=False
    ) -> Dict[str, object]:
        lattices = {}
        for section in self.sections.values():
            lattices.update(
                {
                    section.name: SectionLatticeTranslator.from_section(section).to_xsuite(
                        beam_length=beam_length,
                        env=env,
                        particle_ref=particle_ref,
                        save=save,
                    )
                }
            )
        return lattices


import re
import os
import numpy as np
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from nala.models.element import Magnet
from typing import Any, Dict, Type, get_args, get_origin, Union, Literal

class Counter(dict):
    def __init__(self, sub={}):
        super().__init__()
        self.sub = sub

    def counter(self, type):
        type = self.sub[type] if type in self.sub else type
        if type not in self:
            return 1
        return self[type] + 1

    def value(self, type):
        type = self.sub[type] if type in self.sub else type
        if type not in self:
            return 1
        return self[type]

    def add(self, type, n=1):
        type = self.sub[type] if type in self.sub else type
        if type not in self:
            self[type] = n
        else:
            self[type] += n
        return self[type]

def convert_numpy_types(v):
    if isinstance(v, dict):
        return {key: convert_numpy_types(item) for key, item in v.items()}
    elif isinstance(v, (np.ndarray, list, tuple)):
        try:
            return [convert_numpy_types(li) for li in v]
        except TypeError:
            return float(v)
    elif isinstance(v, (np.float64, np.float32, np.float16)):
        return float(v)
    elif isinstance(
        v,
        (
            np.int_,
            np.intc,
            np.intp,
            np.int8,
            np.int16,
            np.int32,
            np.int64,
            np.uint8,
            np.uint16,
            np.uint32,
            np.uint64,
        ),
    ):
        return int(v)
    elif isinstance(v, field):
        return convert_numpy_types(v.model_dump())
    else:
        return v

def _rotation_matrix(theta):
    return np.array(
        [
            [np.cos(theta), 0, np.sin(theta)],
            [0, 1, 0],
            [-1 * np.sin(theta), 0, np.cos(theta)],
        ]
    )

def chop(expr, delta=1e-8):
    """Performs a chop on small numbers"""
    if isinstance(expr, (int, float, complex)):
        return 0 if -delta <= expr <= delta else expr
    else:
        return [chop(x, delta) for x in expr]

def lattice_to_cartesian(elements):
    """
    Compute Cartesian coordinates [x, y, z] of accelerator lattice elements
    from a sequence of drifts and dipoles in 3D.

    Parameters
    ----------
    elements : list of tuples
        Each element is defined as:
          ("drift", L)
          ("dipole_h", L, phi)   # horizontal bend (x-z plane)
          ("dipole_v", L, phi)   # vertical bend (x-y plane)
        where L = length, phi = bending angle in radians.

    Returns
    -------
    positions : list of tuples
        Cartesian coordinates (x, y, z) for element ends.
    """

    x, y, z = 0.0, 0.0, 0.0  # starting point
    theta_h = 0.0  # azimuth angle in horizontal (x-z)
    theta_v = 0.0  # elevation angle in vertical (x-y)
    positions = [(x, y, z)]

    for elem in elements:
        cond1 = elem.hardware_type.lower() != "dipole"
        cond2 = False
        if isinstance(elem, Magnet):
            if abs(elem.magnetic.angle) > 1e-9:
                cond2 = True
        if cond1 and cond2:
            L = elem.physical.length
            dx = L * np.cos(theta_v) * np.cos(theta_h)
            dy = L * np.sin(theta_v)
            dz = L * np.cos(theta_v) * np.sin(theta_h)
            x, y, z = x + dx, y + dy, z + dz
            positions.append((x, y, z))
        else:  # horizontal bend in x-z plane
            L, phi, tilt = elem.physical.length, elem.magnetic.angle, elem.magnetic.tilt
            if np.isclose(tilt, 0):
                R = L / phi
                cx = x - R * np.sin(theta_h)
                cz = z + R * np.cos(theta_h)
                theta_h_new = theta_h + phi
                x = cx + R * np.sin(theta_h_new)
                z = cz - R * np.cos(theta_h_new)
                theta_h = theta_h_new
            elif np.isclose(tilt, np.pi/2):
                R = L / phi
                cy = y - R * np.sin(theta_v)
                cz = z + R * np.cos(theta_v)
                theta_v_new = theta_v + phi
                y = cy + R * np.sin(theta_v_new)
                z = cz - R * np.cos(theta_v_new)
                theta_v = theta_v_new
            else:
                raise ValueError(f"Unrecognised tilt angle {tilt} for {elem.name}")
            positions.append((x, y, z))
    return positions

def sanitize_kwargs(model_cls: type[BaseModel], data: dict[str, Any]) -> dict[str, Any]:
    sanitized = {}
    for field_name, field in model_cls.model_fields.items():
        value = data.get(field_name, None)
        annotation = field.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)

        if value is None:
            # Allow None if explicitly part of the annotation
            if origin is Union and type(None) in args:
                sanitized[field_name] = None
            continue

        if origin is Union:
            # Handle Union of types (including Optional)
            non_none_args = [arg for arg in args if arg is not type(None)]
            if any(_is_valid_type(value, arg) for arg in non_none_args):
                sanitized[field_name] = value
        elif origin is Literal:
            # Handle Literal values
            if value in args:
                sanitized[field_name] = value
        elif _is_valid_type(value, annotation):
            # Simple direct type match
            sanitized[field_name] = value

    return sanitized

def _is_valid_type(value: Any, annotation: Any) -> bool:
    """
    Helper to check if a value matches a type annotation.
    Handles edge cases like Literal at the leaf level.
    """
    origin = get_origin(annotation)
    if origin is Literal:
        return value in get_args(annotation)
    return isinstance(value, annotation)

def get_field_default(field: FieldInfo) -> Any:
    """Get the default value or instance from a field definition."""
    if callable(field.default_factory):
        try:
            return field.default_factory()
        except Exception:
            return "FACTORY_ERROR"
    if field.default is not None:
        return field.default
    return None

def introspect_model_defaults(
    model_cls: Type[BaseModel],
    flatten: bool = False,
    parent_key: str = "",
    separator: str = "_",
) -> Dict[str, Any]:
    """Recursively introspect a Pydantic model class, extracting default values (including nested)."""
    result = {}

    for field_name, field in model_cls.model_fields.items():
        default_value = get_field_default(field)

        # If the default value is a BaseModel (e.g., from default_factory), recurse
        if isinstance(default_value, BaseModel):
            nested = introspect_model_defaults(
                default_value.__class__,
                flatten=flatten,
                parent_key=f"{parent_key}{separator}{field_name}" if parent_key else field_name,
                separator=separator,
            )
            if flatten:
                result.update(nested)
            else:
                result[field_name] = nested
        else:
            key = f"{parent_key}{separator}{field_name}" if (flatten and parent_key) else field_name
            result[key] = default_value

    return result

def isevaluable(self, s):
    try:
        eval(s)
        return True
    except Exception:
        return False

def path_function(a):
    # a_drive, a_tail = os.path.splitdrive(os.path.abspath(a))
    # b_drive, b_tail = os.path.splitdrive(os.path.abspath(b))
    # if (a_drive == b_drive):
    #     return os.path.relpath(a, b)
    # else:
    if a:
        return os.path.abspath(a)
    return './'


def expand_substitution(self, param, subs={}, elements={}, absolute=False):
    if isinstance(param, str):
        subs["master_lattice_location"] = (
            path_function(self.master_lattice_location)+ "/"
        )
        regex = re.compile(r"\$(.*)\$")
        s = re.search(regex, param)
        if s:
            if isevaluable(self, s.group(1)) is True:
                replaced_str = str(eval(re.sub(regex, str(eval(s.group(1))), param)))
            else:
                replaced_str = re.sub(regex, s.group(1), param)
            for key in subs:
                replaced_str = replaced_str.replace(key, subs[key])
            if os.path.exists(replaced_str):
                replaced_str = path_function(replaced_str).replace("\\", "/")
                # print('\tpath exists', replaced_str)
            for e in elements.keys():
                if e in replaced_str:
                    print("Element is in string!", e, replaced_str)
            return replaced_str
        else:
            return param
    else:
        return param

def checkValue(self, d, default=None):
    if isinstance(d, dict):
        if "type" in d and d["type"] == "list":
            if "default" in d:
                return [
                    a if a is not None else b for a, b in zip(d["value"], d["default"])
                ]
            else:
                if isinstance(d["value"], list):
                    return [val if val is not None else default for val in d["value"]]
                else:
                    return None
        else:
            d["value"] = expand_substitution(self, d["value"])
            return (
                d["value"]
                if d["value"] is not None
                else d["default"] if "default" in d else default
            )
    elif isinstance(d, str):
        return (
            getattr(self, d)
            if hasattr(self, d) and getattr(self, d) is not None
            else default
        )

def tw_cavity_energy_gain(cavity):
    """
    Estimate energy gain in a travelling-wave RF cavity.

    Parameters:
        cavity (nala.models.element.RFCavity): RFCavity element

    Returns:
        float: Estimated energy gain [eV]
    """

    # Approximate effective accelerating gradient
    E_acc = cavity.field_amplitude * np.sin(np.pi * cavity.mode_numerator * 2 / cavity.mode_denominator / 2)

    # Total cavity length
    L_total = cavity.n_cells * cavity.cell_length

    # Energy gain in MeV (since 1 MV/m * 1 m = 1 MeV for charge = e)
    delta_W = E_acc * L_total * np.cos(np.pi * cavity.phase / 180)

    return delta_W



import numpy as np
from scipy.constants import pi, c, e, m_e, epsilon_0
from typing import Any
from warnings import warn
from .base import BaseElementTranslator
from nala.models.plasma import PlasmaElement
from nala.models.simulation import PlasmaSimulationElement
from nala.models.laser import LaserElement
from .laser import LaserTranslator

class PlasmaTranslator(BaseElementTranslator):
    """
    Translator class for converting a :class:`~nala.models.element.Plasma` instance into a string or
    object that can be understood by various simulation codes.
    """

    plasma: PlasmaElement
    """Plasma element."""

    simulation: PlasmaSimulationElement
    """Plasma simulation element."""

    laser: LaserElement | None
    """Laser element."""

    def to_wake_t(self) -> Any:
        """
        Create a Wake-T plasma element object based on the attributes of this element.

        If a `laser` sub-element is defined, it is also converted to a Wake-T laser object and
        added to the plasma element; if two laser sub-elements are defined, they are summed together
        using the :class:`~wake_t.physics_models.laser.laser_pulse.SummedPulse` class.

        Returns
        -------
        wake_t.PlasmaStage
            Wake-T plasma element object

        Raises
        ------
        ValueError
            If the wakefield model is not supported; note that not all models are implemented yet
        """
        from wake_t.physics_models.laser.laser_pulse import (
            SummedPulse,
        )
        from ..conversion_rules.codes import wake_t_conversion

        type_conversion_rules_Wake_T = wake_t_conversion.wake_t_conversion_rules
        if self.simulation.wakefield_model is None:
            warn(
                "No wakefield model defined; no plasma wakefields will be computed."
                 f"Supported models are {list(self.simulation.required_attrs.keys())[1:]}."
            )
        elif self.simulation.wakefield_model not in self.simulation.required_attrs.keys():
            raise ValueError(
                f"Invalid wakefield model {self.wakefield_model}. "
                f"Supported models are {list(self.simulation.required_attrs.keys())[1:]}."
            )
        commondict = {
            self._convertKeyword_WakeT(param): getattr(self, param) for param in
            self.simulation.required_attrs["common"]
        }
        modeldict = {self._convertKeyword_WakeT(param): getattr(self, param) for param in
                     self.simulation.required_attrs[self.simulation.wakefield_model]}
        if self.plasma.density_profile:
            modeldict["density"] = self._density_profile
        else:
            modeldict["density"] = float(self.plasma.density)
        elemdict = modeldict | commondict
        lasers = []
        if isinstance(self.laser, LaserElement):
            laser_translator = LaserTranslator.model_validate(self.model_dump())
            lasers.append(laser_translator.to_wake_t())
        if len(lasers) == 1:
            elemdict.update({"laser": lasers[0]})
        elif len(lasers) == 2:
            elemdict.update({"laser": SummedPulse(lasers[0], lasers[1])})
        elif len(lasers) > 2:
            warn("More than two laser sub-elements found; only the first two will be used.")
            elemdict.update({"laser": SummedPulse(lasers[0], lasers[1])})
        obj = type_conversion_rules_Wake_T[
            self.hardware_type
        ](wakefield_model=self.simulation.wakefield_model, **elemdict)
        return obj

    # Density function.
    def _density_profile(self, z: float) -> np.ndarray:
        """
        Define plasma density as a function of length.
        This takes the :attr:`~length` of the plasma element and calculates the density profile based
        on the :attr:`~ramp_up`, :attr:`~plateau`, :attr:`~ramp_down` and :attr:`~ramp_decay_length` attributes.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps to use for the density profile, by default 1000

        Returns
        -------
        np.ndarray
            Array of length `n_steps` containing the density profile in m^-3
        """
        # Allocate relative density array.
        if self.plasma.plateau <= 0:
            raise ValueError("Plateau length must be positive for density profile.")
        if self.plasma.ramp_up < 0 or self.plasma.ramp_down < 0 or self.plasma.ramp_decay_length <= 0:
            raise ValueError(
                "Ramp lengths must be non-negative and ramp decay length must be positive for density profile."
            )
        n = np.ones_like(z)
        # Add upramp.
        n = np.where(z < self.plasma.ramp_up, 1 / (1 + (self.plasma.ramp_up - z) / self.plasma.ramp_decay_length) ** 2, n)
        # Add downramp.
        try:
            n = np.where(
                (z > self.plasma.ramp_up + self.plasma.plateau) & (z <= self.plasma.ramp_up + self.plasma.plateau + self.plasma.ramp_down),
                1 / (1 + (z - self.plasma.ramp_up - self.plasma.plateau) / self.plasma.ramp_decay_length) ** 2,
                n,
            )
        except ZeroDivisionError:
            n = np.where(
                (z > self.plasma.ramp_up + self.plasma.plateau) & (z <= self.plasma.ramp_up + self.plasma.plateau + self.plasma.ramp_down),
                1,
                n,
            )
        # Make zero after downramp.
        n = np.where(z > self.plasma.ramp_up + self.plasma.plateau + self.plasma.ramp_down, 1e-6, n)
        # Return absolute density.
        return n * self.plasma.density



from pydantic import BaseModel
from typing import Dict, Any, List
from pytao import Tao, TaoCommandError
from collections import Counter
from . import magnetic_orders
import nala.models.element as NALA_elements
from nala.models.elementList import SectionLattice, MachineLayout, ElementList
from nala.models.element import (
    Element,
    Combined_Corrector,
    Vertical_Corrector,
    Horizontal_Corrector,
)
import math
def norm(v):
    n = math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])
    if n == 0:
        return (0.0, 0.0, 0.0)
    return (v[0]/n, v[1]/n, v[2]/n)

def cross(a, b):
    return (a[1]*b[2] - a[2]*b[1],
            a[2]*b[0] - a[0]*b[2],
            a[0]*b[1] - a[1]*b[0])

def dot(a, b):
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]

def deg(r): return math.degrees(r)

def rotation_angles(forward):
    f = norm(forward)
    # yaw (heading around Y): atan2(x, z)
    yaw = math.atan2(f[0], f[2])

    # pitch (elevation): atan2(y, sqrt(x^2+z^2))
    pitch = math.atan2(f[1], math.sqrt(f[0] * f[0] + f[2] * f[2]))

    # build a local right/up so we can compute roll.
    # choose world_up = (0,1,0)
    world_up = (0.0, 1.0, 0.0)

    right = norm(cross(world_up, f))
    up = cross(f, right)  # orthogonal up for the object frame

    # roll = signed angle around forward that rotates object's up -> world_up
    cr = cross(up, world_up)
    roll = math.atan2(dot(cr, f), dot(up, world_up))
    return pitch, roll, yaw

class BmadLatticeImporter(BaseModel):

    floorplan_init: str
    """Name of Tao init file which produces floor coordinates"""

    libtao: str = None
    """libtao.so file"""

    elements: Dict = {}
    """Dictionary containing converted NALA element objects"""

    n_universes: int = 1

    names: Dict[int, Dict[str, List[str]]] = {}

    names_numbered: Dict[int, Dict[str, List[str]]] = {}

    types: Dict[int, Dict[str, List[str]]] = {}

    lengths: Dict[int, Dict[str, List[float]]] = {}

    xpos: Dict[int, Dict[str, List[float]]] = {}

    ypos: Dict[int, Dict[str, List[float]]] = {}

    zpos: Dict[int, Dict[str, List[float]]] = {}

    params: Dict[int, Dict[str, List[Dict[str, Any]]]] = {}

    nala_elems: Dict[int, Dict[str, Dict[str, Element]]] = {}

    branches: Dict[int, List[str]] = {}

    def model_post_init(self, __context: Any) -> None:
        tao = Tao(
            f"-init {self.floorplan_init} -noplot",
            so_lib=self.libtao,
        )
        while True:
            try:
                tao.universe(str(self.n_universes))
                self.branches.update(
                    {
                        self.n_universes: [f'{i["branch_name"]}_{self.n_universes}' for i in
                                           tao.lat_branch_list(ix_uni=self.n_universes)]
                    }
                )
                self.names.update({self.n_universes: {}})
                self.names_numbered.update({self.n_universes: {}})
                self.types.update({self.n_universes: {}})
                self.lengths.update({self.n_universes: {}})
                self.xpos.update({self.n_universes: {}})
                self.ypos.update({self.n_universes: {}})
                self.zpos.update({self.n_universes: {}})
                self.params.update({self.n_universes: {}})
                self.nala_elems.update({self.n_universes: {}})
                for ind, b in enumerate(self.branches[self.n_universes]):
                    kwa = {"ix_uni": str(self.n_universes), "ix_branch": b.replace(f"_{self.n_universes}", "")}
                    names = [i for i in tao.lat_list("*", "ele.name", **kwa)]
                    names_numbered = [f"{x}.{(c := Counter(names[:i + 1]))[x]}" for i, x in enumerate(names)]
                    types = [i for i in tao.lat_list("*", "ele.key", **kwa)]
                    lengths = [i for i in tao.lat_list("*", "ele.l", **kwa)]
                    xpos = [i for i in tao.lat_list("*", "orbit.floor.x", **kwa)]
                    ypos = [i for i in tao.lat_list("*", "orbit.floor.y", **kwa)]
                    zpos = [i for i in tao.lat_list("*", "orbit.floor.z", **kwa)]
                    params = [tao.ele_gen_attribs(f"{str(self.n_universes)}@{ind}>>{i}") for i in range(len(names))]
                    self.names[self.n_universes].update({b: names})
                    self.names_numbered[self.n_universes].update({b: names_numbered})
                    self.types[self.n_universes].update({b: types})
                    self.lengths[self.n_universes].update({b: lengths})
                    self.xpos[self.n_universes].update({b: xpos})
                    self.ypos[self.n_universes].update({b: ypos})
                    self.zpos[self.n_universes].update({b: zpos})
                    self.params[self.n_universes].update({b: params})
                    self.nala_elems[self.n_universes].update({b: {}})
                self.n_universes += 1
            except TaoCommandError:
                break
        self.branches = {
            k: [f'{i["branch_name"]}_{k}' for i in tao.lat_branch_list(ix_uni=k)] for k in
            range(1, self.n_universes)
        }

    def create_element_dictionary(self, universe: int) -> None:
        for b in self.names_numbered[universe].keys():
            for i, nam in enumerate(self.names_numbered[universe][b]):
                end_x = self.xpos[universe][b][i]
                end_y = self.ypos[universe][b][i]
                end_z = self.zpos[universe][b][i]

                # Get the start position (end of previous element, or origin if first)
                if i > 0:
                    start_x = self.xpos[universe][b][i - 1]
                    start_y = self.ypos[universe][b][i - 1]
                    start_z = self.zpos[universe][b][i - 1]
                else:
                    # First element starts at origin (or you might have a specific starting point)
                    start_x = 0.0
                    start_y = 0.0
                    start_z = 0.0

                # Calculate middle position as the average of start and end
                middle = [
                    (start_x + end_x) / 2.0,
                    (start_y + end_y) / 2.0,
                    (start_z + end_z) / 2.0
                ]

                # Calculate the beam direction at the START of this element
                # This is the direction the beam is traveling when it enters the element
                if i > 1:
                    # Direction from the element before last to the last element
                    # This gives us the exit direction of the previous element
                    forward = (
                        self.xpos[universe][b][i - 1] - self.xpos[universe][b][i - 2],
                        self.ypos[universe][b][i - 1] - self.ypos[universe][b][i - 2],
                        self.zpos[universe][b][i - 1] - self.zpos[universe][b][i - 2]
                    )
                elif i == 1:
                    # For the second element, use direction from origin to first element's end
                    forward = (
                        self.xpos[universe][b][0],
                        self.ypos[universe][b][0],
                        self.zpos[universe][b][0]
                    )
                else:
                    # For the first element, assume initial direction along z-axis
                    # (or you might have a specific initial direction)
                    forward = (0.0, 0.0, 1.0)

                # Calculate rotation angles from the forward direction at element start
                pitch, roll, yaw = rotation_angles(forward)

                elem_data = {}
                parameters = self.params[universe][b][i]
                if self.types[universe][b][i] == "Kicker":
                    hardware_class = "Magnet"
                    hardware_type = "Combined_Corrector"
                    horizontal = nam + "_H"
                    vertical = nam + "_V"
                    hcor = {
                        "magnetic": {
                            "order": 0,
                            "length": self.lengths[universe][b][i],
                            "kl": parameters["HKICK"],
                        }
                    }
                    vcor = {
                        "magnetic": {
                            "order": 0,
                            "length": float(self.lengths[universe][b][i]),
                            "kl": parameters["VKICK"],
                        }
                    }
                    elem_data = {
                        "hardware_type": hardware_type,
                        "Horizontal_Corrector": horizontal,
                        "Vertical_Corrector": vertical,
                        "hcor": hcor,
                        "vcor": vcor,
                    }
                elif self.types[universe][b][i] in magnetic_orders:
                    hardware_type = self.types[universe][b][i]
                    hardware_class = "Magnet"
                    try:
                        kl = {
                            "multipoles": {
                                f"K{magnetic_orders[hardware_type]}L": {
                                    "normal": parameters[f"K{magnetic_orders[hardware_type]}"] * self.lengths[universe][b][i],
                                    "order": magnetic_orders[hardware_type],
                                },
                            },
                        }
                    except KeyError:
                        kl = {
                            "multipoles": {
                                    f"K{magnetic_orders[hardware_type]}L": {
                                        "normal": parameters["ANGLE"],
                                        "order": magnetic_orders[hardware_type],
                                    },
                                },
                            "entrance_edge_angle": parameters["E1"],
                            "exit_edge_angle": parameters["E2"],
                        }
                    if "GAP" in parameters:
                        kl.update({"GAP": parameters["GAP"]})
                    elem_data = {
                        "hardware_type": hardware_type,
                        "magnetic": {
                            "order": magnetic_orders[hardware_type],
                            "length": float(self.lengths[universe][b][i]),
                            **kl,
                        }
                    }
                elif self.types[universe][b][i] == "Marker":
                    hardware_class = "Simulation"
                    hardware_type = "Marker"
                    markelem = {
                        "physical": {
                                    "position": middle,
                                    "global_rotation": [pitch, roll, yaw, ],
                                    "length": float(self.lengths[universe][b][i]),
                                },
                        "name": nam,
                        "hardware_class": hardware_class,
                        "hardware_type": hardware_type,
                        "machine_area": "test",
                    }
                    self.nala_elems[universe][b].update(
                        {
                            nam: getattr(NALA_elements, self.types[universe][b][i])(**markelem)
                        }
                    )
                if elem_data:
                    elems = {
                            nam: {
                                "physical": {
                                    "position": middle,
                                    "global_rotation": [pitch, roll, yaw, ],
                                    "length": float(self.lengths[universe][b][i]),
                                },
                                "name": nam,
                                "hardware_class": hardware_class,
                                "hardware_type": hardware_type,
                                "machine_area": "test",
                            }
                        }
                    elems[nam].update(**elem_data)
                    if self.types[universe][b][i] == "Kicker":
                        helem = elems.copy()
                        helem.update(
                            {
                                "name": horizontal,
                                "hardware_type": "Horizontal_Corrector",
                                "magnetic": hcor,
                            }
                        )
                        velem = elems.copy()
                        velem.update(
                            {
                                "name": vertical,
                                "hardware_type": "Vertical_Corrector",
                                "magnetic": vcor,
                            }
                        )
                        comb = Combined_Corrector(**elems[nam])
                        hori = Horizontal_Corrector(**elems[nam])
                        vert = Vertical_Corrector(**elems[nam])
                        self.nala_elems[universe][b].update(
                            {
                                nam: comb,
                                horizontal: hori,
                                vertical: vert,
                            },
                        )
                    elif self.types[universe][b][i] in magnetic_orders:
                        if self.types[universe][b][i] in ["RBend", "SBend"]:
                            self.types[universe][b][i] = "Dipole"
                            elems[nam]["hardware_type"] = "Dipole"
                            elems[nam]["physical"]["physical_angle"] = -elems[nam]["magnetic"]["multipoles"]["K0L"]["normal"]
                        self.nala_elems[universe][b].update(
                            {
                                nam: getattr(NALA_elements, self.types[universe][b][i])(**elems[nam])
                            }
                        )

    def create_section(self, universe: int, branch: str) -> Dict[str, SectionLattice]:
        if not self.elements:
            self.create_element_dictionary(universe)
        order = list(self.nala_elems[universe][branch].keys())
        elems = self.nala_elems[universe][branch]
        seclat = SectionLattice(order=order, elements=ElementList(elements=elems), name=branch)
        return {branch: seclat}

    def create_layout(self, universe: int) -> Dict[str, MachineLayout]:
        layout = {}
        for branch in list(self.names_numbered[universe].keys()):
            layout.update(**self.create_section(universe, branch))
        return {str(universe): MachineLayout(name=str(universe), sections=layout)}



from xtrack.beam_elements import Solenoid as Solenoid_xs
from xtrack.beam_elements import Bend as Bend_xs
from xtrack.beam_elements import DipoleEdge as DipoleEdge_xs
from xtrack.beam_elements import Quadrupole as Quadrupole_xs
from xtrack.beam_elements import Sextupole as Sextupole_xs
from xtrack.beam_elements import Octupole as Octupole_xs
from xtrack.beam_elements import Drift as Drift_xs
from xtrack.beam_elements import NonLinearLens as NonLinearLens_xs
from xtrack.beam_elements import Cavity as Cavity_xs
from xtrack.beam_elements import UniformSolenoid as UniformSolenoid_xs
from xtrack.beam_elements import Multipole as Multipole_xs
from xtrack.beam_elements import Marker as Marker_xs
from xtrack.monitors import ParticlesMonitor as ParticlesMonitor_xs

from nala.models.element import (
    Dipole,
    Solenoid,
    Quadrupole,
    Sextupole,
    Octupole,
    RFCavity,
    Drift,
    Solenoid,
    NonLinearLens,
    Magnet,
    Marker,
)

xsuite_conversion_rules_reverse = {
    Bend_xs: Dipole,
    DipoleEdge_xs: Marker,
    Solenoid_xs: Solenoid,
    Quadrupole_xs: Quadrupole,
    Sextupole_xs: Sextupole,
    Octupole_xs: Octupole,
    Cavity_xs: RFCavity,
    Drift_xs: Drift,
    UniformSolenoid_xs: Solenoid,
    NonLinearLens_xs: NonLinearLens,
    Multipole_xs: Magnet,
    Marker_xs: Marker,
}

xsuite_conversion_rules = {
    "Dipole": Bend_xs,
    "Solenoid": Solenoid_xs,
    "Quadrupole": Quadrupole_xs,
    "Sextupole": Sextupole_xs,
    "Octupole": Octupole_xs,
    "Beam_Position_Monitor": ParticlesMonitor_xs,
    "Beam_Arrival_Monitor": Drift_xs,
    "Bunch_Length_Monitor": Drift_xs,
    "Screen": ParticlesMonitor_xs,
    "Marker": ParticlesMonitor_xs,
    "Rcollimator": Drift_xs,
    "Collimator": Drift_xs,
    "Monitor": Marker_xs,
    "Wall_Current_Monitor": Drift_xs,
    "Integrated_Current_Transformer": Drift_xs,
    "Faraday_Cup": Drift_xs,
    "RFCavity": Cavity_xs,
    "RFDeflectingCavity": Cavity_xs,
    "Aperture": Drift_xs,
    "Shutter": Drift_xs,
    "Valve": Drift_xs,
    "Bellows": Drift_xs,
    "Cleaner": Drift_xs,
    "Drift": Drift_xs,
    "NonLinearLens": NonLinearLens_xs,
    "Combined_Corrector": Bend_xs,
    "Horizontal_Corrector": Bend_xs,
    "Vertical_Corrector": Bend_xs,
    "Scatter": Marker_xs,
    "APContour": Marker_xs,
    "Center": Marker_xs,
    "FEL_Modulator": Drift_xs,
    "Wiggler": Drift_xs,
    "Charge": Marker_xs,
    "Wakefield": Marker_xs,
    "Watch_Point": ParticlesMonitor_xs,
    "Laser": Drift_xs,
}


import numpy as np
from collections import OrderedDict
from pydantic import create_model, BaseModel
from yaml import safe_load


def _rotation_matrix(theta):
    return np.array(
        [
            [np.cos(theta), 0, np.sin(theta)],
            [0, 1, 0],
            [-1 * np.sin(theta), 0, np.cos(theta)],
        ]
    )


def read_yaml(fname: str) -> BaseModel:
    with open(fname, "r") as f:
        data = safe_load(f)

    # Build fields: field_name: (type, default)
    fields = {key: (type(value), value) for key, value in data.items()}

    # Create and return the dynamic model class
    DynamicModel = create_model(
        "DynamicModel",
        __base__=BaseModel,
        __module__=__name__,
        # model_config=model_config,
        **fields,
    )
    return DynamicModel(**data)


def merge_two_dicts(y, x):
    """Combine to dictionaries: first dictionary overwrites keys in the second dictionary"""
    if not isinstance(x, (dict, OrderedDict)) and not isinstance(
        y, (dict, OrderedDict)
    ):
        return OrderedDict()
    elif not isinstance(x, (dict, OrderedDict)):
        return y
    elif not isinstance(y, (dict, OrderedDict)):
        return x
    else:
        z = x.copy()  # start with x's keys and values
        z.update(y)  # modifies z with y's keys and values & returns None
        return z



"""
NALA Element Module

The main class for representing accelerator elements in NALA.
"""
import os
from typing import Type, List, Union, Dict, Tuple, Any, get_args, get_origin
from pydantic import field_validator, Field, BaseModel
import types
from .control import ControlsInformation

from .baseModels import T, Aliases, IgnoreExtra
from .manufacturer import ManufacturerElement
from .electrical import ElectricalElement
from .degauss import DegaussableElement
from .physical import PhysicalElement, Rotation
from .reference import ReferenceElement
from .magnetic import (
    MagneticElement,
    Dipole_Magnet,
    Quadrupole_Magnet,
    Sextupole_Magnet,
    Octupole_Magnet,
    Solenoid_Magnet,
    NonLinearLens_Magnet,
    Wiggler_Magnet,
)
from .plasma import PlasmaElement
from .diagnostic import (
    Beam_Position_Monitor_Diagnostic,
    Beam_Arrival_Monitor_Diagnostic,
    Bunch_Length_Monitor_Diagnostic,
    Camera_Diagnostic,
    Screen_Diagnostic,
    Charge_Diagnostic,
)
from .laser import LaserElement, LaserEnergyMeterElement, LaserMirrorElement, LaserHalfWavePlateElement
from .lighting import LightingElement
from .RF import (
    PIDElement,
    Low_Level_RF_Element,
    RFModulatorElement,
    RFProtectionElement,
    RFHeartbeatElement,
    RFCavityElement,
    RFDeflectingCavityElement,
    WakefieldElement,
)
from .shutter import ShutterElement, ValveElement
from .simulation import (
    ApertureElement,
    RFCavitySimulationElement,
    WakefieldSimulationElement,
    MagnetSimulationElement,
    DriftSimulationElement,
    DiagnosticSimulationElement,
    PlasmaSimulationElement,
    SimulationElement,
    TwissMatchSimulationElement,
)
import yaml
from collections.abc import MutableMapping


def flatten(dictionary: Dict, parent_key: str="", separator: str="_") -> Dict:
    """
    Flatten a nested dictionary -- used for expanding the nested `BaseModel` structure.

    Args:
        dictionary (Dict): The dictionary to flatten.
        parent_key (str, optional): The base key to use for the flattened keys. Defaults to "".
        separator (str, optional): The separator to use between keys. Defaults to "_".

    Returns:
        Dict: The flattened dictionary.
    """
    items = []
    for key, value in dictionary.items():
        if isinstance(key, str):
            new_key = parent_key + separator + key if parent_key else key
            if isinstance(value, MutableMapping):
                items.extend(flatten(value, new_key, separator=separator).items())
            else:
                items.append((new_key, value))
    return dict(items)


class string_with_quotes(str):
    pass


class flow_list(list):
    pass


def flow_list_rep(dumper, data):
    return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)


def quoted_presenter(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style='"')


yaml.add_representer(string_with_quotes, quoted_presenter)
yaml.add_representer(flow_list, flow_list_rep)


class baseElement(IgnoreExtra):
    """
    Base-level element class. All NALA elements derive from this.

    Attributes:
        name (str): The name of the element.
        hardware_class (str): The hardware class of the element.
        hardware_type (str): The hardware type of the element.
        hardware_model (str): The hardware model of the element.
        machine_area (str): The machine area of the element.
        virtual_name (str): The virtual name of the element.
        alias (Aliases): The alias(es) of the element.
        subelement (bool | str): Whether the element is a subelement.
    """

    name: str
    """Name of the element."""

    hardware_class: str
    """Hardware class of the element."""

    hardware_type: str
    """Hardware type of the element."""

    hardware_model: str = Field(default="Generic", frozen=True)
    """Specific model of the element."""

    machine_area: str
    """Machine are of the element."""

    virtual_name: str = ""
    """Name of the element in the virtual control system."""

    alias: Union[str, list, Aliases, None] = Field(alias="name_alias", default=None)
    """The alias(es) of the element"""

    subelement: bool | str = False
    """Flag to indicate whether the element is a subelement of another 
    (i.e. whether they overlap in physical space)."""

    # Define cascading rules: (source_path, target_path)
    CASCADING_RULES: Dict = {}

    @field_validator("name", mode="before")
    @classmethod
    def validate_name(cls, v: str) -> str:
        assert isinstance(v, str)
        # try:
        #     PV(pv=str(v) + ":")
        # except Exception:
        #     raise ValueError("name is not a valid element name")
        return v

    @field_validator("alias", mode="before")
    @classmethod
    def validate_alias(cls, v: Union[str, List, None]) -> Aliases:
        # print(list(map(str.strip, v.split(','))))
        if isinstance(v, str):
            return Aliases(aliases=list(map(str.strip, v.split(","))))
        elif isinstance(v, (list, tuple)):
            return Aliases(aliases=list(v))
        elif isinstance(v, (dict)):
            return Aliases(aliases=v["aliases"])
        elif v is None:
            return Aliases(aliases=[])
        else:
            raise ValueError("alias should be a string or a list of strings")

    def escape_string_list(self, escapes) -> str:
        if len(list(escapes)) > 0:
            return string_with_quotes(",".join(map(str, list(escapes))))
        return string_with_quotes("")

    @classmethod
    def from_CATAP(cls: Type[T], fields: dict) -> T:
        return cls(**fields)

    # def generate_aliases(self) -> list:
    #     magnetPV = PV.fromString(str(self.name) + ":")  # ('CLA', 'S07', 'QUAD', 1)
    #     return [
    #         magnetPV.area + "-" + magnetPV.typename + str(magnetPV.index).zfill(2),
    #         magnetPV.area + "-" + magnetPV.typename + str(magnetPV._indexString),
    #         magnetPV.area + "-" + magnetPV.typename + str(magnetPV.index),
    #     ]

    def _resolve_attribute_path(self, attr_name: str) -> List[Tuple[str, ...]]:
        """
        Helper to get the full path(s) for a given attribute name.
        """
        return self._find_field_paths(attr_name, self.__class__)

    @classmethod
    def _find_field_paths(
            cls: Type['main'],
            attr_name: str,
            current_model: Type[BaseModel],
            current_path: Tuple[str, ...] = ()
    ) -> List[Tuple[str, ...]]:
        """
        Recursively searches for an attribute name within the model's structure
        and returns a list of its full access paths.
        """
        paths = []
        for field_name, field_info in current_model.model_fields.items():
            new_path = current_path + (field_name,)

            # 1. Check if the current field is the target attribute
            if field_name == attr_name:
                paths.append(new_path)

            # 2. Check if the field is a nested Pydantic model (recursive step)
            field_annotation = field_info.annotation
            origin = get_origin(field_annotation)

            # Unwrap Union/Optional types to get the actual class
            is_union = origin is Union or isinstance(field_annotation, types.UnionType)
            if is_union:
                args = get_args(field_annotation)
                for arg in args:
                    if arg is not type(None):
                        try:
                            if isinstance(arg, type) and issubclass(arg, BaseModel):
                                paths.extend(cls._find_field_paths(attr_name, arg, new_path))
                        except TypeError:
                            # arg might not be a class, skip it
                            pass
            else:
                try:
                    if isinstance(field_annotation, type) and issubclass(field_annotation, BaseModel):
                        paths.extend(cls._find_field_paths(attr_name, field_annotation, new_path))
                except TypeError:
                    pass

        for name, attr in vars(current_model).items():
            if isinstance(attr, property) and name == attr_name:
                paths.append(current_path + (name,))

        return paths

    def _get_nested_attribute(self, path: Tuple[str, ...]) -> Any:
        """
        Accesses a nested attribute using its path.
        """
        value = self
        for step in path:
            if value is None:
                raise AttributeError(f"Cannot access '{step}' on None value at path '{'.'.join(path)}'")
            value = getattr(value, step)
        return value

    def _set_nested_attribute(self, path: Tuple[str, ...], value: Any):
        """
        Sets a nested attribute using its path.
        """
        target_model = self

        # Traverse to the second to last element
        for step in path[:-1]:
            target_model = getattr(target_model, step)
            if target_model is None:
                raise AttributeError(f"Cannot set attribute at path '{'.'.join(path)}': intermediate value is None")

        # Set the attribute
        setattr(target_model, path[-1], value)

    def __getattr__(self, name: str) -> Any:
        """
        Custom getter: Looks for the attribute in nested models.
        """
        # Avoid recursion on special attributes
        if name.startswith('_'):
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

        paths = self._resolve_attribute_path(name)

        if not paths:
            raise AttributeError(f"'{self.__class__.__name__}' object and its nested models have no attribute '{name}'")

        if len(paths) > 1:
            path_strings = [f"'{'.'.join(p)}'" for p in paths]
            raise AttributeError(
                f"Attribute '{name}' is ambiguous. Found at: {', '.join(path_strings)}. "
                "Access explicitly (e.g., `element.simulation.field_amplitude`)."
            )

        return self._get_nested_attribute(paths[0])

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Custom setter with cascading updates for related attributes.
        """
        cls = self.__class__

        # Allow Pydantic to handle direct fields and internal attributes
        if name in cls.model_fields or name.startswith('_'):
            super().__setattr__(name, value)
            return

        # Try nested lookup
        try:
            paths = self._resolve_attribute_path(name)
        except Exception:
            super().__setattr__(name, value)
            return

        if not paths:
            super().__setattr__(name, value)
            return

        if len(paths) > 1:
            path_strings = [f"'{'.'.join(p)}'" for p in paths]
            raise AttributeError(
                f"Cannot set ambiguous attribute '{name}'. Found at: {', '.join(path_strings)}. "
                "Set explicitly."
            )

        # Set the nested attribute
        self._set_nested_attribute(paths[0], value)

        # Handle cascading updates
        self._handle_cascading_updates(paths[0], value)

    def _handle_cascading_updates(self, path: Tuple[str, ...], value: Any) -> None:
        """
        Handle cascading attribute updates across nested models.
        """

        for source_path, target_path in self.CASCADING_RULES.items():
            if path == source_path:
                self._set_nested_attribute(target_path, value)

    def to_CATAP(self) -> dict:
        return {
            "machine_area": self.machine_area,
            "hardware_type": self.hardware_type,
            "name": self.name,
            # 'virtual_name': self.virtual_name,
            "name_alias": (
                self.alias.aliases if isinstance(self.alias, Aliases) else self.alias
            ),
        }

    @property
    def no_controls(self) -> str:
        cls = self.__class__
        return (
            self.__class__.__name__
            + "("
            + " ".join(
                [
                    k + "=" + getattr(self, k).__repr__()
                    for k in cls.model_fields.keys()
                    if k != "controls"
                ]
            )
            + ")"
        )

    @property
    def subdirectory(self) -> str:
        if self.__class__.__name__ == self.hardware_type:
            return os.path.join(self.hardware_class, self.hardware_type)
        return os.path.join(
            self.hardware_class, self.__class__.__name__, self.hardware_type
        )

    @property
    def YAML_filename(self) -> str:
        return os.path.join(self.subdirectory, self.name + ".yaml")

    @property
    def hardware_info(self) -> Dict[str, str]:
        """
        Retrieve the `hardware_class` and `hardware_type` of the object as a dict

        Returns
        -------
            Dict[str, str]: {"class": `hardware_class`, "type": `hardware_type`}
        """
        return {"class": self.hardware_class, "type": self.hardware_type}

    def flat(self) -> Dict[str, Any]:
        """
        Dump the entire element model as a flat dictionary, with sub-models separated by "_".

        For example, if an element has `element.electrical.maxI` this will be keyed in the dictionary as
        `eletrical_maxI: value`

        Returns
        -------
            Dict[str, Any]: Flattened dictionary representing the element.
        """
        return flatten(self.model_dump(), parent_key="", separator="_")

    def is_subelement(self) -> bool:
        """
        Flag to indicate whether the element is a subelement of another, such as a solenoid surrounding
        an RF cavity or a BPM embedded inside a magnet. This precludes it from being included when calculating
        the full length of a beamline.

        Returns
        -------
            bool: True if the element is a subelement.
        """
        if str(self.subelement).lower() == "false":
            return False
        elif str(self.subelement).lower() == "true":
            return True
        if isinstance(self.subelement, bool):
            return self.subelement
        else:
            return isinstance(self.subelement, str)


class Element(baseElement):
    """
    Standard class for representing elements.

    Attributes:
        simulation: :class:`~nala.models.simulation.SimulationElement`: The simulation attributes of the element.
        electrical: :class:`~nala.models.electrical.ElectricalElement`: The electrical attributes of the element.
        manufacturer: :class:`~nala.models.manufacturer.Manufacturer`: The manufacturer attributes of the element.
        controls: :class:`~nala.models.control.ControlsInformation` | None: The control system attributes of the element.
        reference: :class:`~nala.models.reference.ReferenceElement` | None: Reference information for the element.
    """

    simulation: SimulationElement = Field(default_factory=SimulationElement)
    """Simulation attributes of the element."""

    electrical: ElectricalElement | None = Field(default_factory=ElectricalElement)
    """Electrical attributes of the element."""

    manufacturer: ManufacturerElement | None = Field(default_factory=ManufacturerElement)
    """Manufacturer attributes of the element."""

    controls: ControlsInformation | None = None
    """Control system attributes of the element."""

    reference: ReferenceElement | None = None
    """Additional reference information for the element."""

    def to_CATAP(self):
        catap_dict = super().to_CATAP()
        catap_dict.update(
            {
                "manufacturer": self.manufacturer.manufacturer,
                "serial_number": self.manufacturer.serial_number,
            }
        )
        return catap_dict

class PhysicalBaseElement(Element):
    """
    Element with a physical attribute; see :class:`~nala.models.physical.PhysicalElement`.

    Attributes:
        physical: PhysicalElement: The physical attributes of the element.
    """

    physical: PhysicalElement = Field(default_factory=PhysicalElement)
    """Physical attributes of the element."""

    def to_CATAP(self):
        catap_dict = super().to_CATAP()
        catap_dict.update(
            {
                "position": list(self.physical.middle)[2],
            }
        )
        return catap_dict

    @property
    def bend_angle(self) -> Rotation:
        """
        Bending angle of the element.
        #TODO this probably doesn't do what it should.

        Returns
        -------
            :class:`~nala.models.physical.Rotation`: The rotation attribute of the element.
        """
        return Rotation.from_list([0, 0, 0])

    @property
    def start_angle(self) -> Rotation:
        """
        Initial global rotation angle of the element.
        #TODO this probably doesn't do what it should.
        """
        return self.physical.rotation + self.physical.global_rotation

    @property
    def end_angle(self) -> Rotation:
        """
        Final global rotation angle of the element
        #TODO this probably doesn't do what it should.
        """
        return self.start_angle



class Magnet(PhysicalBaseElement):
    """
    Base class for representing magnets.

    Attributes:
        degauss: :class:`~nala.models.degauss.DegaussableElement`: The degaussing attributes of the magnet.
        simulation: :class:`~nala.models.simulation.MagnetSimulationElement`: The simulation attributes of the magnet.
        magnetic: :class:`~nala.models.magnetic.MagneticElement` | None: The magnetic attributes of the magnet.
    """

    hardware_class: str = Field(default="Magnet", frozen=True)
    """Magnet hardware class."""

    degauss: DegaussableElement | None = Field(default_factory=DegaussableElement)
    """Degaussing attributes of the magnet."""

    simulation: MagnetSimulationElement = Field(default_factory=MagnetSimulationElement)
    """Simulation attributes of the magnet."""

    magnetic: MagneticElement | None = None
    """Magnetic attributes of the magnet."""

    @property
    def bend_angle(self) -> Rotation:
        """
        Rotation of the magnet based on its bending angle.
        """
        return Rotation.from_list([0, 0, self.magnetic.angle])

    @property
    def end_angle(self) -> float:
        """End angle of the magnet"""
        return self.start_angle + self.bend_angle.theta

    # @field_validator('type', mode='before')
    # @classmethod
    # def validate_type(cls, v: str) -> str:
    #     # print(list(map(str.strip, v.split(','))))
    #     if isinstance(v, str):
    #         return v.upper()
    #     else:
    #         raise ValueError('alias should be a string or a list of strings')

    def to_CATAP(self):
        catap_dict = super().to_CATAP()
        catap_dict.update(
            {
                "mag_type": self.hardware_type,
                "degauss_tolerance": self.degauss.tolerance,
                "degauss_values": self.escape_string_list(self.degauss.values),
                "num_degauss_steps": self.degauss.steps,
                "field_integral_coefficients": self.escape_string_list(
                    self.magnetic.field_integral_coefficients
                ),
                "linear_saturation_coefficients": self.escape_string_list(
                    self.magnetic.linear_saturation_coefficients
                ),
                "mag_set_max_wait_time": self.magnetic.settle_time,
                "magnetic_length": 1000 * self.magnetic.length,
                "ri_tolerance": self.electrical.read_tolerance,
                "min_i": self.electrical.minI,
                "max_i": self.electrical.maxI,
            }
        )
        return catap_dict

    # @property
    # def subdirectory(self):
    #     return os.path.join(self.hardware_type,self.type)


class Dipole(Magnet):
    """
    Dipole element.

    Attributes:
        hardware_type (str): The hardware type of the dipole.
        magnetic (:class:`~nala.models.magnetic.Dipole_Magnet`): The magnetic attributes of the dipole.
    """

    hardware_type: str = Field(default="Dipole", frozen=True)
    """Dipole hardware type."""

    magnetic: Dipole_Magnet = Field(default_factory=Dipole_Magnet)


class Quadrupole(Magnet):
    """
    Quadrupole element.

    Attributes:
        hardware_type (str): The hardware type of the quadrupole.
        magnetic (:class:`~nala.models.magnetic.Quadrupole_Magnet`): The magnetic attributes of the quadrupole.
    """

    hardware_type: str = Field(default="Quadrupole", frozen=True)
    """Quadrupole hardware type."""

    magnetic: Quadrupole_Magnet = Field(default_factory=Quadrupole_Magnet)
    """Magnetic attributes of the quadrupole."""



class Sextupole(Magnet):
    """
    Sextupole element.

    Attributes:
        hardware_type (str): The hardware type of the sextupole.
        magnetic (:class:`~nala.models.magnetic.Sextupole_Magnet`): The magnetic attributes of the sextupole.
    """

    hardware_type: str = Field(default="Sextupole", frozen=True)
    """Sextupole hardware type."""

    magnetic: Sextupole_Magnet = Field(default_factory=Sextupole_Magnet)
    """Magnetic attributes of the sextupole."""


class Octupole(Magnet):
    """
    Octupole element.

    Attributes:
        hardware_type (str): The hardware type of the octupole.
        magnetic (:class:`~nala.models.magnetic.Octupole_Magnet`): The magnetic attributes of the octupole.
    """

    hardware_type: str = Field(default="Octupole", frozen=True)
    """Octupole hardware type."""

    magnetic: Octupole_Magnet = Field(default_factory=Octupole_Magnet)
    """Magnetic attributes of the octupole."""


class Horizontal_Corrector(Dipole):
    """
    Horizontal corrector element.

    Attributes:
        hardware_type (str): The hardware type of the corrector.
    """

    hardware_type: str = Field(default="Horizontal_Corrector", frozen=True)
    """Horizontal corrector hardware type."""


class Vertical_Corrector(Dipole):
    """
    Vertical corrector element.

    Attributes:
        hardware_type (str): The hardware type of the corrector.
    """

    hardware_type: str = Field(default="Vertical_Corrector", frozen=True)
    """Vertical corrector hardware type."""


class Combined_Corrector(Dipole):
    """
    Horizontal corrector element.

    Attributes:
        hardware_type (str): The hardware type of the corrector.
        Horizontal_Corrector (str): The horizontal corrector.
        Vertical_Corrector (str): The vertical corrector.
    """

    hardware_type: str = Field(default="Combined_Corrector", frozen=True)
    """Combined corrector hardware type."""

    Horizontal_Corrector: str | None = Field(default=None, frozen=True)
    """Name of horizontal corrector."""

    Vertical_Corrector: str | None = Field(default=None, frozen=True)
    """Name of vertical corrector."""


class Solenoid(Magnet):
    """
    Solenoid element.

    Attributes:
        hardware_type (str): The hardware type of the solenoid.
        magnetic (:class:`~nala.models.magnetic.Solenoid_Magnet`): The magnetic attributes of the solenoid.
    """

    hardware_type: str = Field(default="Solenoid", frozen=True)
    """Solenoid hardware type."""

    magnetic: Solenoid_Magnet = Field(default_factory=Solenoid_Magnet)
    """Magnetic attributes of the solenoid."""


class NonLinearLens(Magnet):
    """
    Non-linear lens element.

    Attributes:
        hardware_type (str): The hardware type of the NLL.
        magnetic (:class:`~nala.models.magnetic.NonLinearLens_Magnet`): The magnetic attributes of the NLL.
    """

    hardware_type: str = Field(default="NonLinearLens", frozen=True)
    """Non-linear lens hardware type."""

    magnetic: NonLinearLens_Magnet = Field(default_factory=NonLinearLens_Magnet)
    """Magnetic attributes of the non-linear-lens."""


class Wiggler(Magnet):
    """
    Wiggler element.

    Attributes:
        hardware_type (str): The hardware type of the wiggler.
        magnetic (:class:`~nala.models.magnetic.Wiggler_Magnet`): The magnetic attributes of the wiggler.
        laser (:class:`~nala.models.laser.Laser_Magnet` or None): The laser associated with the wiggler.
    """

    hardware_type: str = Field(default="Undulator", frozen=True)
    """Wiggler hardware type."""

    magnetic: Wiggler_Magnet = Field(default_factory=Wiggler_Magnet)
    """Magnetic attributes of the wiggler."""

    laser: LaserElement | None = None
    """Laser attached to the wiggler."""


class TwissMatch(PhysicalBaseElement):
    """
    Twiss matching element. Used for changing the Twiss parameters of the beam.

    Attributes:
        hardware_type (str): The hardware type of the element.
        hardware_class (str): The hardware class of the element.
        simulation (:class:`~nala.models.simulation.TwissMatchSimulationElement`): The simulation attributes of the matching element.
    """

    hardware_type: str = Field(default="TwissMatch", frozen=True)
    """Twiss match hardware type."""

    hardware_class: str = Field(default="TwissMatch", frozen=True)
    """Twiss match hardware class."""

    simulation: TwissMatchSimulationElement = Field(default_factory=TwissMatchSimulationElement)
    """Simulation attributes of the matching element."""


class Diagnostic(PhysicalBaseElement):
    """
    Base class for representing diagnostics.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        hardware_class (str): The hardware class of the diagnostic.
        simulation: (:class:`~nala.models.simulation.DiagnosticSimulationElement`): The simulation
        attributes of the diagnostic (including its `output_filename`).
    """

    hardware_type: str = Field(default="Diagnostic", frozen=True)
    """Diagnostic hardware type."""

    hardware_class: str = Field(default="Diagnostic", frozen=True)
    """Diagnostic hardware class."""

    simulation: DiagnosticSimulationElement = Field(default_factory=DiagnosticSimulationElement)
    """Simulation attributes of the diagnostic."""


class Beam_Position_Monitor(Diagnostic):
    """
    BPM element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        hardware_model (str): The specific hardware model of the diagnostic (i.e. Stripline, Cavity).
        diagnostic: (:class:`~nala.models.diagnostic.Beam_Position_Monitor_Diagnostic`): The diagnostic
        attributes of the BPM.
    """

    hardware_type: str = Field(default="Beam_Position_Monitor", frozen=True, alias="BPM")
    """BPM hardware type."""

    hardware_model: str = Field(default="Stripline", frozen=True)
    """BPM hardware model."""

    diagnostic: Beam_Position_Monitor_Diagnostic = Field(default_factory=Beam_Position_Monitor_Diagnostic)
    """Diagnostic attributes of the BPM."""


class Beam_Arrival_Monitor(Diagnostic):
    """
    BAM element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        hardware_model (str): The specific hardware model of the diagnostic.
        diagnostic: (:class:`~nala.models.diagnostic.Beam_Arrival_Monitor_Diagnostic`): The diagnostic
        attributes of the BAM.
    """

    hardware_type: str = Field(default="Beam_Arrival_Monitor", frozen=True, alias="BAM")
    """BAM hardware type."""

    hardware_model: str = Field(default="DESY", frozen=True)
    """BAM hardware model."""

    diagnostic: Beam_Arrival_Monitor_Diagnostic = Field(default_factory=Beam_Arrival_Monitor_Diagnostic)
    """Diagnostic attributes of the BAM."""


class Bunch_Length_Monitor(Diagnostic):
    """
    BLM element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        hardware_model (str): The specific hardware model of the diagnostic.
        diagnostic: (:class:`~nala.models.diagnostic.Bunch_Length_Monitor_Diagnostic`): The diagnostic
        attributes of the BLM.
    """

    hardware_type: str = Field(default="Bunch_Length_Monitor", frozen=True, alias="BLM")
    """BLM hardware type."""

    hardware_model: str = Field(default="CDR", frozen=True)
    """BLM hardware model."""

    diagnostic: Bunch_Length_Monitor_Diagnostic = Field(default_factory=Bunch_Length_Monitor_Diagnostic)
    """Diagnostic attributes of the BLM."""


class Camera(Diagnostic):
    """
    Camera element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        hardware_model (str): The specific hardware model of the diagnostic.
        diagnostic: (:class:`~nala.models.diagnostic.Camera_Diagnostic`): The diagnostic
        attributes of the Camera.
    """

    hardware_type: str = Field(default="Camera", frozen=True)
    """Camera hardware type."""

    hardware_model: str = Field(default="PCO", frozen=True)
    """Camera hardware model."""

    diagnostic: Camera_Diagnostic = Field(default_factory=Camera_Diagnostic)
    """Diagnostic attributes of the camera."""


class Screen(Diagnostic):
    """
    Screen element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        hardware_model (str): The specific hardware model of the diagnostic.
        diagnostic: (:class:`~nala.models.diagnostic.Screen_Diagnostic`): The diagnostic
        attributes of the Screen.
    """

    hardware_type: str = Field(default="Screen", frozen=True)
    """Screen hardware type."""

    hardware_model: str = Field(default="YAG", frozen=True)
    """Screen hardware model."""

    diagnostic: Screen_Diagnostic = Field(default_factory=Screen_Diagnostic)
    """Diagnostic attributes of the screen."""

    def to_CATAP(self):
        catap_dict = super().to_CATAP()
        catap_dict.update(
            {
                "screen_type": self.diagnostic.type,
                "has_camera": self.diagnostic.has_camera,
                "camera_name": self.diagnostic.camera_name,
                "devices": self.escape_string_list(self.diagnostic.devices),
            }
        )
        return catap_dict


class ChargeDiagnostic(Diagnostic):
    """
    Generic charge diagnostic element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
        diagnostic: (:class:`~nala.models.diagnostic.Charge_Diagnostic`): The diagnostic
        attributes of the diagnostic.
    """

    hardware_type: str = Field(default="ChargeDiagnostic", frozen=True)
    """Charge diagnostic hardware type."""

    diagnostic: Charge_Diagnostic = Field(default_factory=Charge_Diagnostic)
    """Diagnostic attributes of the charge diagnostic."""


class Wall_Current_Monitor(ChargeDiagnostic):
    """
    WCM charge diagnostic element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
    """

    hardware_type: str = Field(default="Wall_Current_Monitor", frozen=True, alias="WCM")
    """WCM hardware type."""


class Faraday_Cup_Monitor(ChargeDiagnostic):
    """
    FCM charge diagnostic element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
    """

    hardware_type: str = Field(default="Faraday_Cup_Monitor", frozen=True, alias="FCM")
    """FCM hardware type."""


class Integrated_Current_Transformer(ChargeDiagnostic):
    """
    ICT charge diagnostic element.

    Attributes:
        hardware_type (str): The hardware type of the diagnostic.
    """

    hardware_type: str = Field(default="Integrated_Current_Transformer", frozen=True, alias="ICT")
    """ICT hardware type."""


class VacuumGauge(PhysicalBaseElement):
    """
    Vacuum gauge element.

    Attributes:
        hardware_type (str): The hardware type of the gauge.
        hardware_model (str): The hardware model of the gauge.
    """

    hardware_type: str = Field(default="VacuumGauge", frozen=True)
    """Vacuum gauge hardware type."""

    hardware_model: str = Field(default="IMG", frozen=True)
    """Vacuum gauge hardware model."""


class Laser(PhysicalBaseElement):
    """
    Laser element.

    Attributes:
        hardware_type (str): The hardware type of the laser.
        hardware_model (str): The hardware model of the laser.
        laser (:class:`~nala.models.laser.LaserElement`): The laser attributes of the laser.
    """

    hardware_type: str = Field(default="Laser", frozen=True)
    """Laser hardware type."""

    hardware_model: str = Field(default="Laser", frozen=True)
    """Laser hardware model."""

    laser: LaserElement = Field(default_factory=LaserElement)
    """Laser attributes of the laser."""


class LaserEnergyMeter(Element):
    """
    Laser energy meter element.

    Attributes:
        hardware_type (str): The hardware type of the laser energy meter.
        hardware_model (str): The hardware model of the laser energy meter.
        laser (:class:`~nala.models.laser.LaserEnergyMeterElement`): The laser-related attributes of the
        energy meter.
    """

    hardware_type: str = Field(default="LaserEnergyMeter", frozen=True)
    """Laser energy meter hardware type."""

    hardware_model: str = Field(default="Gentec Photodiode", frozen=True)
    """Laser energy meter hardware model.
    #TODO should be manufacturer?"""

    laser: LaserEnergyMeterElement = Field(default_factory=LaserEnergyMeterElement)
    """Laser energy meter attributes of the element."""


class LaserHalfWavePlate(Element):
    """
    Laser half-wave plate element.

    Attributes:
        hardware_type (str): The hardware type of the HWP.
        hardware_model (str): The hardware model of the HWP.
        laser (:class:`~nala.models.laser.LaserHalfWavePlateElement`): The laser-related attributes of the
        HWP.
    """

    hardware_type: str = Field(default="LaserHalfWavePlate", frozen=True)
    """Laser half-wave plate element type."""

    hardware_model: str = Field(default="Newport", frozen=True)
    """Laser half-wave plate hardware model.
    #TODO should be manufacturer?"""

    laser: LaserHalfWavePlateElement = Field(default_factory=LaserHalfWavePlateElement)
    """Laser half-wave plate element attributes of the element."""


class LaserMirror(Element):
    """
    Laser mirror element.

    Attributes:
        hardware_type (str): The hardware type of the mirror.
        hardware_model (str): The hardware model of the mirror.
        laser (:class:`~nala.models.laser.LaserMirrorElement`): The laser-related attributes of the
        mirror.
    """

    hardware_type: str = Field(default="LaserMirror", frozen=True)
    """Laser mirror hardware type."""

    hardware_model: str = Field(default="Planar", frozen=True)
    """Laser mirror hardware model."""

    laser: LaserMirrorElement = Field(default_factory=LaserMirrorElement)
    """Laser mirror attributes of the element."""


class Plasma(PhysicalBaseElement):
    """
    Plasma element.

    Attributes:
        hardware_type (str): The hardware type of the element.
        simulation (:class:`~nala.models.simulation.PlasmaSimulationElement`): The simulation
        attributes of the plasma.
        plasma (:class:`~nala.models.plasma.PlasmaElement`): The plasma attributes of the plasma.
        laser (:class:`~nala.models.laser.LaserElement` or None): The laser assosicated with the plasma
    """

    hardware_type: str = Field(default="Plasma", frozen=True)
    """Plasma hardware type."""

    simulation: PlasmaSimulationElement = Field(default_factory=PlasmaSimulationElement)
    """Simulation attributes of the plasma."""

    plasma: PlasmaElement = Field(default_factory=PlasmaElement)
    """Plasma attribute of the plasma."""

    laser: LaserElement | None = None
    """Laser attached to the plasma element."""


class Lighting(Element):
    """
    Lighting element.

    Attributes:
        hardware_type (str): The hardware type of the element.
        hardware_model (str): The hardware model of the element.
        lights (:class:`~nala.models.lighting.LightingElement`): The lighting element.
    """

    hardware_type: str = Field(default="Lighting", frozen=True)
    """Lighting hardware type."""

    hardware_model: str = Field(default="LED", frozen=True)
    """Lighting hardware model."""

    lights: LightingElement = Field(default_factory=LightingElement)
    """Lighting attributes of the element."""


class PID(Element):
    """
    Proportional-integral-derivative feedback element.

    Attributes:
        hardware_type (str): The hardware type of the element.
        hardware_model (str): The hardware model of the element.
        PID (:class:`~nala.models.RF.PIDElement`): The PID element.
    """

    hardware_type: str = Field(default="PID", frozen=True)
    """PID hardware type."""

    hardware_model: str = Field(default="RF", frozen=True)
    """PID hardware model."""

    PID: PIDElement = Field(default_factory=PIDElement)
    """PID attributes of the element."""


class Low_Level_RF(Element):
    """
    Low-level RF element.

    Attributes:
        hardware_type (str): The hardware type of the element.
        hardware_model (str): The hardware model of the element.
        LLRF (:class:`~nala.models.RF.Low_Level_RF_Element`): The LLRF element.
    """

    hardware_type: str = Field(default="Low_Level_RF", frozen=True)
    """LLRF hardware type."""

    hardware_model: str = Field(default="Libera", frozen=True)
    """LLRF hardware model."""

    LLRF: Low_Level_RF_Element = Field(default_factory=Low_Level_RF_Element)
    """LLRF attributes of the element."""


class RFCavity(PhysicalBaseElement):
    """
    RFCavity element.

    Attributes:
        hardware_type (str): The hardware type of the RF cavity.
        hardware_model (str): The specific hardware model of the RF cavity.
        cavity (:class:`~nala.models.RF.RFCavityElement`): The RF cavity attributes of the element.
        simulation: (:class:`~nala.models.simulation.RFCavitySimulationElement`): The simulation
        attributes of the RF cavity.
    """

    hardware_type: str = Field(default="RFCavity", frozen=True)
    """RF cavity hardware type."""

    hardware_model: str = Field(default="SBand", frozen=True)
    """RF cavity hardware model."""

    cavity: RFCavityElement = Field(default_factory=RFCavityElement)
    """Cavity attributes of the RF cavity."""

    simulation: RFCavitySimulationElement = Field(default_factory=RFCavitySimulationElement)
    """Simulation attributes of the RF cavity."""


class Wakefield(PhysicalBaseElement):
    """
    Wakefield element.

    Attributes:
        hardware_type (str): The hardware type of the wakefield.
        hardware_model (str): The specific hardware model of the wakefield.
        cavity: (:class:`~nala.models.RF.WakefieldElement`): The wakefield cavity attributes of the element.
        simulation: (:class:`~nala.models.simulation.WakefieldSimulationElement`): The simulation
        attributes of the wakefield cavity.
    """

    hardware_type: str = Field(default="Wakefield", frozen=True)
    """Wakefield hardware type."""

    hardware_model: str = Field(default="Dielectric", frozen=True)
    """Wakefield hardware model."""

    cavity: WakefieldElement = Field(default_factory=WakefieldElement)
    """Wakefield attributes of the element."""

    simulation: WakefieldSimulationElement = Field(default_factory=WakefieldSimulationElement)
    """Simulation attributes of the wakefield element."""


class RFDeflectingCavity(RFCavity):
    """
    RF Deflecting Cavity element.

    Attributes:
        hardware_type (str): The hardware type of the RF cavity.
        hardware_model (str): The specific hardware model of the RF cavity.
        cavity (:class:`~nala.models.RF.RFDeflectingCavityElement`): The RF cavity attributes of the element.
        simulation: (:class:`~nala.models.simulation.RFCavitySimulationElement`): The simulation
        attributes of the RF cavity.
    """

    hardware_type: str = Field(default="RFDeflectingCavity", frozen=True)
    """RF deflecting cavity hardware type."""

    hardware_model: str = Field(default="SBand", frozen=True)
    """RF deflecting cavity hardware model."""

    cavity: RFDeflectingCavityElement = Field(default_factory=RFDeflectingCavityElement)
    """Cavity attributes of the RF deflecting cavity."""

    simulation: RFCavitySimulationElement = Field(default_factory=RFCavitySimulationElement)
    """Simulation attributes of the RF deflecting cavity."""


class RFModulator(Element):
    """
    RF Modulator element.

    Attributes:
        hardware_type (str): The hardware type of the RF modulator.
        hardware_model (str): The specific hardware model of the RF modulator.
        modulator (:class:`~nala.models.RF.RFModulatorElement`): The RF modulator attributes of the element.
    """

    hardware_type: str = Field(default="RFModulator", frozen=True)
    """RF modulator hardware type."""

    hardware_model: str = Field(default="Thales", frozen=True)
    """RF modulator hardware model.
    #TODO move to manufacturer?"""

    modulator: RFModulatorElement = Field(default_factory=RFModulatorElement)
    """RF modulator attributes of the element."""


class RFProtection(Element):
    """
    RF Protection element.

    Attributes:
        hardware_type (str): The hardware type of the RF protection system.
        hardware_model (str): The specific hardware model of the RF protection system.
        modulator (:class:`~nala.models.RF.RFProtectionElement`): The RF protection attributes of the element.
    """

    hardware_type: str = Field(default="RFProtection", frozen=True)
    """RF protection hardware type."""

    hardware_model: str = Field(default="PROT", frozen=True)
    """RF protection hardware model."""

    modulator: RFProtectionElement = Field(default_factory=RFProtectionElement)
    """RF protection attributes of the element."""


class RFHeartbeat(Element):
    """
    RF Heartbeat element.

    Attributes:
        hardware_type (str): The hardware type of the RF heartbeat system.
        heartbeat (:class:`~nala.models.RF.RFHeartbeatElement`): The RF heartbeat attributes of the element.
    """

    hardware_type: str = Field(default="RFHeartbeat", frozen=True)
    """RF heartbeat hardware type."""

    heartbeat: RFHeartbeatElement = Field(default_factory=RFHeartbeatElement)
    """RF heartbeat system attributes."""


class Shutter(PhysicalBaseElement):
    """
    Shutter element.

    Attributes:
        hardware_type (str): The hardware type of the shutter.
        shutter (:class:`~nala.models.shutter.ShutterElement`): The shutter attributes of the element.
    """

    hardware_type: str = Field(default="Shutter", frozen=True)
    """Shutter hardware type"""

    shutter: ShutterElement = Field(default_factory=ShutterElement)
    """Shutter attributes of the element."""


class Valve(PhysicalBaseElement):
    """
    Vacuum valve element.

    Attributes:
        hardware_type (str): The hardware type of the valve.
        valve (:class:`~nala.models.shutter.ValveElement`): The valve attributes of the element.
    """

    hardware_type: str = Field(default="Valve", frozen=True)
    """Valve hardware type."""

    valve: ValveElement = Field(default_factory=ValveElement)
    """Valve attributes of the element."""


class Marker(PhysicalBaseElement):
    """
    Marker element.

    Attributes:
        hardware_type (str): The hardware type of the marker.
        hardware_model (str): The hardware model of the marker.
        simulation (:class:`~nala.models.simulation.DiagnosticSimulationElement`): The simulation
        attributes of the marker.
    """

    hardware_type: str = Field(default="Marker", frozen=True)
    """Marker hardware type."""

    hardware_model: str = Field(default="Simulation", frozen=True)
    """Marker hardware model."""

    simulation: DiagnosticSimulationElement = Field(default_factory=DiagnosticSimulationElement)
    """Simulation attributes of the marker."""


class Aperture(PhysicalBaseElement):
    """
    Aperture element.

    Attributes:
        hardware_type (str): The hardware type of the aperture.
        hardware_model (str): The hardware model of the aperture.
        aperture (:class:`~nala.models.simulation.ApertureElement`): The simulation
        attributes of the aperture.
    """

    hardware_type: str = Field(default="Aperture", frozen=True)
    """Aperture hardware type."""

    hardware_model: str = Field(default="Simulation", frozen=True)
    """Aperture hardware model."""

    aperture: ApertureElement = Field(default_factory=ApertureElement)
    """Aperture attributes of the element."""


class Collimator(Aperture):
    """
    Collimator element.

    Attributes:
        hardware_type (str): The hardware type of the collimator.
        hardware_model (str): The hardware model of the collimator.
    """

    hardware_type: str = Field(default="Collimator", frozen=True)
    """Collimator hardware type."""

    hardware_model: str = Field(default="Simulation", frozen=True)
    """Collimator hardware model."""


class Drift(PhysicalBaseElement):
    """
    Drift element.

    Attributes:
        hardware_type (str): The hardware type of the marker.
        simulation (:class:`~nala.models.simulation.DriftSimulationElement`): The simulation
        attributes of the drift.
    """

    hardware_type: str = Field(default="Drift", frozen=True)
    """Drift hardware type."""

    simulation: DriftSimulationElement = Field(default_factory=DriftSimulationElement)
    """Simulation attributes of the drift."""



from yaml.constructor import SafeConstructor


# Create custom safe constructor class that inherits from SafeConstructor
class MySafeConstructor(SafeConstructor):

    # Create new method handle boolean logic
    def add_bool(self, node):
        return self.construct_scalar(node)


# Inject the above boolean logic into the custom constuctor
MySafeConstructor.add_constructor("tag:yaml.org,2002:bool", MySafeConstructor.add_bool)



import sys

sys.path.append("..")

from pprint import pprint  # noqa E402
from PAdantic.models.magnetic import Quadrupole_Magnet  # noqa E402
from PAdantic.models.physical import PhysicalElement, Position  # noqa E402
from PAdantic.models.element import Quadrupole  # noqa E402
from PAdantic.models.elementList import MachineModel  # noqa E402
from PAdantic.PAdantic import PAdantic  # noqa E402

# We are able to specify layouts even without section definitions
machine = MachineModel(
    layout={
        "layouts": {
            "line1": {
                "FODO": ["QUAD1"],
                "NODO": ["QUAD2"],
            },
        },
    },
)

# Make an element from scratch and add it to the machine
q1 = Quadrupole(
    name="QUAD1",
    machine_area="FODO",
    magnetic=Quadrupole_Magnet(
        length=0.1,
        k1l=1.0,
    ),
    physical=PhysicalElement(
        middle=Position(
            x=0,
            y=0,
            z=0.1,
        ),
        length=0.1,
    ),
)
machine.append({q1.name: q1})
print(f"Machine sections after QUAD-01 addition: {machine.sections}")

# Make another element from scratch and add it to the machine
q2 = Quadrupole(
    name="QUAD2",
    machine_area="NODO",
    magnetic=Quadrupole_Magnet(
        length=0.1,
        k1l=1.0,
    ),
    physical=PhysicalElement(
        middle=Position(
            x=0,
            y=0,
            z=0.5,
        ),
        length=0.1,
    ),
)
machine.append({q2.name: q2})

print("============= Machine Model Creation =============")
# Check the sections and layouts have been built correctly
print(f"Machine sections after QUAD-02 addition: {machine.sections}")
print(f"Machine lattices after QUAD-02 addition: {machine.lattices}")

sections = {
    "sections": {
        "FODO": ["QUAD1"],
        "NODO": ["QUAD2"],
    }
}
layouts = {
    "layouts": {
        "line1": ["FODO", "NODO"],
    }
}

print("============= PADantic Creation =============")
pa_machine = PAdantic(element_list=[q1, q2], layout=layouts, section=sections)
pprint(f"PADantic creation result [elements]: {pa_machine.elements}")
print(f"PADantic creation result [sections]: {pa_machine.sections}")
print(f"PADantic creation result [lattices]: {pa_machine.lattices}")
pa_machine.get_magnets()



import unittest
from nala.models.control import ControlVariable, ControlsInformation


class TestControlVariable(unittest.TestCase):
    def test_control_variable_creation(self):
        cv = ControlVariable(
            identifier="var1",
            dtype="float",
            protocol="CA",
            units="V",
            description="A float variable for voltage",
        )
        self.assertEqual(cv.identifier, "var1")
        self.assertEqual(cv.dtype, float)
        self.assertEqual(cv.protocol, "CA")
        self.assertEqual(cv.units, "V")
        self.assertEqual(cv.description, "A float variable for voltage")

    def test_dtype_as_type(self):
        cv = ControlVariable(
            identifier="var1",
            dtype=float,
            protocol="CA",
        )
        self.assertEqual(cv.dtype, float)

    def test_validation_for_missing_dtype(self):
        with self.assertRaises(ValueError):
            ControlVariable(
                identifier="var1",
                protocol="CA",
            )

    def test_validation_for_missing_identifier(self):
        with self.assertRaises(ValueError):
            ControlVariable(
                dtype="float",
                protocol="CA",
            )

    def test_validation_for_missing_protocol(self):
        with self.assertRaises(ValueError):
            ControlVariable(
                identifier="var1",
                dtype="float",
            )

    def test_invalid_dtype(self):
        with self.assertRaises(ValueError):
            ControlVariable(
                identifier="var2",
                dtype="unknown_type",
                protocol="PVA",
            )


class TestControlsInformation(unittest.TestCase):
    def test_controls_information_creation(self):
        controls_info = ControlsInformation(
            variables={
                "var1": ControlVariable(
                    identifier="var1",
                    dtype="float",
                    protocol="CA",
                    units="V",
                    description="A float variable for voltage",
                ),
                "var2": ControlVariable(
                    identifier="var2",
                    dtype="int",
                    protocol="PVA",
                ),
            }
        )
        self.assertIn("var1", controls_info.variables)
        self.assertIn("var2", controls_info.variables)
        self.assertEqual(controls_info.variables["var1"].dtype, float)
        self.assertEqual(controls_info.variables["var2"].dtype, int)

    def test_controls_information_with_dicts(self):
        controls_info = ControlsInformation(
            variables={
                "var1": {
                    "identifier": "var1",
                    "dtype": "float",
                    "protocol": "CA",
                    "units": "V",
                    "description": "A float variable for voltage",
                },
                "var2": {
                    "identifier": "var2",
                    "dtype": "int",
                    "protocol": "PVA",
                },
            }
        )
        self.assertIn("var1", controls_info.variables)
        self.assertIn("var2", controls_info.variables)
        self.assertEqual(controls_info.variables["var1"].dtype, float)
        self.assertEqual(controls_info.variables["var2"].dtype, int)

    def test_controls_information_with_mixed_types(self):
        controls_info = ControlsInformation(
            variables={
                "var1": ControlVariable(
                    identifier="var1",
                    dtype="float",
                    protocol="CA",
                    units="V",
                    description="A float variable for voltage",
                ),
                "var2": {
                    "identifier": "var2",
                    "dtype": "int",
                    "protocol": "PVA",
                },
            }
        )
        self.assertIn("var1", controls_info.variables)
        self.assertIn("var2", controls_info.variables)
        self.assertEqual(controls_info.variables["var1"].dtype, float)
        self.assertEqual(controls_info.variables["var2"].dtype, int)

    def test_controls_information_with_invalid_dict(self):
        with self.assertRaises(ValueError):
            ControlsInformation(
                variables={
                    "var1": {
                        "identifier": "var1",
                        "dtype": "float",
                        # Missing protocol
                    },
                }
            )

    def test_controls_information_with_invalid_type(self):
        with self.assertRaises(TypeError):
            ControlsInformation(variables=["not", "a", "dict"])

    def test_controls_information_model_dump_serialises_dtypes(self):
        controls_info = ControlsInformation(
            variables={
                "var1": ControlVariable(
                    identifier="var1",
                    dtype=float,
                    protocol="CA",
                    units="V",
                    description="A float variable for voltage",
                ),
                "var2": ControlVariable(
                    identifier="var2",
                    dtype=int,
                    protocol="PVA",
                ),
            }
        )
        dumped = controls_info.model_dump()
        self.assertEqual(dumped["variables"]["var1"]["dtype"], "float")
        self.assertEqual(dumped["variables"]["var2"]["dtype"], "int")



import pytest
from copy import deepcopy

from nala.unit_tests.models.test_element import physical_base_element
from nala.models.elementList import SectionLattice, ElementList, MachineLayout, MachineModel
from nala.models.exceptions import LatticeError

def chunks(li, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(li), n):
        yield li[i: i + n]

@pytest.fixture
def section_lattice(physical_base_element):
    be1 = deepcopy(physical_base_element)
    be1.name = "elem1"
    be1.physical.middle.z = 1.0
    be1.physical.length = 0.1
    be2 = deepcopy(physical_base_element)
    be2.name = "elem2"
    be2.physical.middle.z = 2.0
    be2.physical.length = 0.1
    elements = ElementList(elements={"elem1": be1, "elem2": be2})
    return SectionLattice(name="TestSection", order=["elem1", "elem2"], elements=elements)

@pytest.fixture
def machine_layout(section_lattice):
    return MachineLayout(name="TestLayout", sections={"TestSection": section_lattice})

@pytest.fixture
def machine_model(section_lattice, machine_layout):
    return MachineModel(
        elements=section_lattice.elements.elements,
        section={"sections": {section_lattice.name: section_lattice.names}},
        layout={
            "layouts": {
                machine_layout.name: list(machine_layout.sections.keys())
            },
            "default_layout": machine_layout.name,
        },
    )

def test_section_lattice_names(section_lattice):
    assert section_lattice.names == ["elem1", "elem2"]

def test_section_lattice_create_drifts(section_lattice):
    drifts = section_lattice.createDrifts()
    print(drifts)
    assert isinstance(drifts, dict)

def test_section_lattice_get_s_values(section_lattice):
    s_values = section_lattice.get_s_values(as_dict=True)
    print(s_values)
    assert isinstance(s_values, dict)
    s_values = section_lattice.get_s_values()
    assert isinstance(s_values, list)

def test_machine_layout_names(machine_layout):
    assert machine_layout.names == ["TestSection"]

def test_machine_layout_get_element(machine_layout):
    with pytest.raises(LatticeError):
        machine_layout.get_element("nonexistent")

def test_machine_layout_elements_between(machine_layout):
    elements = machine_layout.elements_between()
    assert isinstance(elements, list)

def test_machine_model_addition(machine_model, physical_base_element):
    new_elements = {"elem3": physical_base_element}
    result = machine_model + new_elements
    assert "elem3" in result

def test_machine_model_get_element(machine_model):
    element = machine_model.get_element("elem1")
    assert element.name == "elem1"

def test_machine_model_elements_between(machine_model):
    elements = machine_model.elements_between()
    assert isinstance(elements, list)


from typing import List
import numpy as np
import easygdf
from warnings import warn
from .FieldParameter import FieldParameter
from ..units import UnitValue
from scipy.constants import speed_of_light


def write_gdf_field_file(self) -> str:
    """
    Generate the field data in a format that is suitable for GPT, based on the
    :class:`~nala.translator.utils.fields.field` object provided.
    This is then written to a GDF file.
    The `field_type` parameter determines the format of the file.

    A warning is raised if the field type is not supported (perhaps elevate to a `NotImplementedError`?)

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object

    Returns
    -------
    str:
        The name of the GDF field file.
    """
    gdf_file = self._output_filename(extension=".gdf")
    blocks = None
    zdata = self.z.value.val
    if self.field_type == "LongitudinalWake":
        wzdata = self.Wz.value.val
        blocks = union(
            [
                {"name": "z", "value": zdata},
                {"name": "Wz", "value": wzdata},
            ]
        )
    elif self.field_type == "TransverseWake":
        wxdata = self.Wx.value.val
        wydata = self.Wy.value.val
        blocks = [
            {"name": "z", "value": zdata},
            {"name": "Wx", "value": wxdata},
            {"name": "Wy", "value": wydata},
        ]
    elif self.field_type == "3DWake":
        wxdata = self.Wx.value.val
        wydata = self.Wy.value.val
        wzdata = self.Wz.value.val
        blocks = [
            {"name": "z", "value": zdata},
            {"name": "Wx", "value": wxdata},
            {"name": "Wy", "value": wydata},
            {"name": "Wz", "value": wzdata},
        ]
    elif self.field_type == "1DMagnetoStatic":
        bzdata = self.Bz.value.val
        blocks = union(
            [
                {"name": "z", "value": zdata},
                {"name": "Bz", "value": bzdata},
            ]
        )
    elif self.field_type == "3DMagnetoStatic":
        xdata = self.x.value.val
        ydata = self.y.value.val
        bxdata = self.Bx.value.val
        bydata = self.By.value.val
        bzdata = self.Bz.value.val
        blocks = [
            {"name": "x", "value": xdata},
            {"name": "y", "value": ydata},
            {"name": "z", "value": zdata},
            {"name": "Bx", "value": bxdata},
            {"name": "By", "value": bydata},
            {"name": "Bz", "value": bzdata},
        ]
    elif self.field_type == "1DElectroDynamic":
        ezdata = self.Ez.value.val
        fielddata = np.array([zdata, ezdata]).transpose()
        if self.cavity_type == "TravellingWave":
            startpos = list(zdata).index(self.start_cell_z)
            # stoppos = list(zdata).index(self.end_cell_z)
            halfcell1 = 1.0 * fielddata[:startpos]
            halfcell2 = 1.0 * halfcell1[::-1]
            halfcell1[:, 1] /= max(halfcell1[:, 1])
            halfcell2[:, 0] = halfcell2[:, 0][::-1]
            halfcell2[:, 1] /= max(halfcell2[:, 1])
            halfcell1end = halfcell1[-1, 0]
            zstep = zdata[1] - zdata[0]
            lambdaRF = speed_of_light / self.frequency
            ncells = (self.n_cells - 1) * self.mode_numerator / self.mode_denominator
            nsteps = int(np.floor((ncells) * lambdaRF / zstep))
            middleRF = np.array(
                [
                    [
                        (x * zstep + halfcell1end + zstep),
                        np.cos((2 * np.pi / lambdaRF) * (x * zstep)),
                    ]
                    for x in range(0, nsteps + 1)
                ]
            )
            halfcell2[:, 0] += middleRF[-1, 0] + zstep
            zdata, ezdata = np.concatenate([halfcell1, middleRF, halfcell2]).transpose()
        blocks = union(
            [
                {"name": "z", "value": zdata},
                {"name": "Ez", "value": ezdata},
            ]
        )
    else:
        warn(f"Field type {self.field_type} not supported for GPT")
    if blocks is not None:
        # print(blocks)
        # print(union(blocks))
        easygdf.save(gdf_file, blocks)
    return gdf_file


def union(blocks: List) -> List:
    """
    Update the field data into a format compatible for easyGDF

    Parameters
    ----------
    blocks: List[Dict]
        The field parameters, keyed by name

    Returns
    -------
    List:
        A list of easyGDF-compatible dictionaries
    """
    names = [b["name"] for b in blocks]
    if "z" in names:
        zidx = names.index("z")
        _, indices = np.unique(
            np.round(blocks[zidx]["value"], decimals=6), return_index=True
        )
        blocks = [{"name": b["name"], "value": b["value"][indices]} for b in blocks]
        return blocks
    return blocks


def read_gdf_field_file(
    self,
    filename: str,
    field_type: str,
    cavity_type: str | None = None,
    frequency: float | None = None,
    normalize_b: bool = True
):
    """
    Read a GDF field file and convert it into a :class:`SimulationFramework.Modules.Fields.field` object

    Parameters
    ----------
    self: :class:`~SimulationFramework.Modules.Fields.field`
        The field object to be updated.
    filename: str
        The path to the GDF field file
    field_type: str
        The name of the field, see :attr:`~SimulationFramework.Modules.Fields.allowed_fields`
    cavity_type: str, optional
        The type of RF cavity, see :attr:`~SimulationFramework.Modules.Fields.allowed_cavities`
    frequency: float, optional
        The frequency of the RF cavity.
    normalize_b: bool, optional
        Normalize Bx and By with respect to Bz (True by default)

    Returns
    -------
    None

    Raises
    ------
    ValueError:
        if the cavity `field_type` contains the string `Electro` and `cavity_type` is not provided
    ValueError:
        if the cavity `field_type` contains the string `Electro` and `frequency` is not provided
    NotImplementedError:
        if a given `field_type` is not implemented
    """
    self.reset_dicts()
    setattr(self, "field_type", field_type)
    if "Electro" in field_type:
        if cavity_type is None:
            raise ValueError(f"cavity_type must be provided for {field_type}")
        else:
            setattr(self, "cavity_type", cavity_type)
        if frequency is None:
            raise ValueError(f"frequency must be provided for {field_type}")
        else:
            setattr(self, "frequency", frequency)
    fdat = easygdf.load(filename)["blocks"]
    try:
        zval = [k["value"] for k in fdat if k["name"].lower() == "z"][0]
    except Exception:
        zval = [k["value"] * speed_of_light for k in fdat if k["name"].lower() == "t"][
            0
        ]
    if field_type == "1DMagnetoStatic":
        bzval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Bz"][0]
        setattr(self, "z", FieldParameter(name="z", value=UnitValue(zval, units="m")))
        setattr(
            self,
            "Bz",
            FieldParameter(
                name="Bz", value=UnitValue(bzval / np.max(bzval), units="T")
            ),
        )
    elif field_type == "3DMagnetoStatic":
        xval = [k["value"] for k in fdat if k["name"].lower() == "x"][0]
        yval = [k["value"] for k in fdat if k["name"].lower() == "y"][0]
        bxval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Bx"][0]
        byval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "By"][0]
        bzval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Bz"][0]
        # Normalise by the maximum *on-axis* Bz field
        if normalize_b:
            normBz = max(
                [abs(Bz) for x, y, Bz in zip(xval, yval, bzval) if x == 0.0 and y == 0.0]
            )
        else:
            normBz = 1
        setattr(self, "x", FieldParameter(name="x", value=UnitValue(xval, units="m")))
        setattr(self, "y", FieldParameter(name="y", value=UnitValue(yval, units="m")))
        setattr(self, "z", FieldParameter(name="z", value=UnitValue(zval, units="m")))
        setattr(
            self,
            "Bx",
            FieldParameter(name="Bx", value=UnitValue(bxval / normBz, units="T")),
        )
        setattr(
            self,
            "By",
            FieldParameter(name="By", value=UnitValue(byval / normBz, units="T")),
        )
        setattr(
            self,
            "Bz",
            FieldParameter(name="Bz", value=UnitValue(bzval / normBz, units="T")),
        )
    elif field_type == "1DElectroDynamic":
        if cavity_type == "StandingWave":
            ezval = [
                k["value"] for k in fdat if k["name"].lower().capitalize() == "Ez"
            ][0]
            setattr(
                self, "z", FieldParameter(name="z", value=UnitValue(zval, units="m"))
            )
            setattr(
                self,
                "Ez",
                FieldParameter(
                    name="Ez", value=UnitValue(ezval / np.max(ezval), units="V/m")
                ),
            )
        elif cavity_type == "TravellingWave":
            raise NotImplementedError(f"{cavity_type} not implemented for GDF files")
    elif field_type == "LongitudinalWake":
        wzval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Wz"][0]
        setattr(self, "z", FieldParameter(name="z", value=UnitValue(zval, units="m")))
        setattr(
            self, "Wz", FieldParameter(name="Wz", value=UnitValue(wzval, units="V/C"))
        )
    elif field_type == "TransverseWake":
        wxval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Wx"][0]
        wyval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Wy"][0]
        setattr(self, "z", FieldParameter(name="z", value=UnitValue(zval, units="m")))
        setattr(
            self, "Wx", FieldParameter(name="Wx", value=UnitValue(wxval, units="V/C/m"))
        )
        setattr(
            self, "Wy", FieldParameter(name="Wy", value=UnitValue(wyval, units="V/C/m"))
        )
    elif field_type == "3DWake":
        wxval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Wx"][0]
        wyval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Wy"][0]
        wzval = [k["value"] for k in fdat if k["name"].lower().capitalize() == "Wz"][0]
        setattr(self, "z", FieldParameter(name="z", value=UnitValue(zval, units="m")))
        setattr(
            self, "Wx", FieldParameter(name="Wx", value=UnitValue(wxval, units="V/C/m"))
        )
        setattr(
            self, "Wy", FieldParameter(name="Wy", value=UnitValue(wyval, units="V/C/m"))
        )
        setattr(
            self, "Wz", FieldParameter(name="Wz", value=UnitValue(wzval, units="V/C"))
        )
    else:
        raise NotImplementedError(f"{field_type} loading not implemented for GDF files")



import os
import h5py
from .FieldParameter import FieldParameter
from ..units import UnitValue
from warnings import warn


allowed_cavities = [
    "StandingWave",
    "TravellingWave",
]

tw_required_attrs = [
    "start_cell_z",
    "end_cell_z",
    "mode_numerator",
    "mode_denominator",
]


def read_HDF5_field_file(self, filename: str) -> str:
    """
    Read an HDF5 field file and convert it into a :class:`nala.translator.utils.fields.field` object

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object.
    filename: str
        The path to the HDF5 field file

    Returns
    -------
    str:
        The name of the HDF5 file

    Raises
    ------
    KeyError:
        if the `cavity_type` attribute is not in :attr:`~nala.translator.utils.fields.allowed_cavities
    KeyError:
        if `cavity_type==TravellingWave` and all the required attributes are not defined in the file;
         see :attr:`~nala.translator.utils.fields.tw_required_attrs
    Warning:
        if a given parameter in the field is not normalised to 1.0
    """
    self.reset_dicts()
    with h5py.File(filename, "r") as h5file:
        for key, value in h5file.attrs.items():
            if key == "type":
                setattr(self, "field_type", value)
            else:
                setattr(self, key, value)
        if "origin_code" in list(h5file.attrs.keys()):
            self.origin_code = h5file.attrs["origin_code"]
        if "norm" in list(h5file.attrs.keys()):
            if h5file.attrs["norm"] != 1.0:
                warn("Warning: Field is not normalised to 1.0.")
            self.norm = h5file.attrs["norm"]
        else:
            self.norm = 1.0
        if "cavity_type" in list(h5file.attrs.keys()):
            cavtype = h5file.attrs["cavity_type"]
            if cavtype not in allowed_cavities:
                raise KeyError(
                    f"cavity_type attributes of {filename} must be in {allowed_cavities}"
                )
            else:
                setattr(self, "cavity_type", cavtype)
            if cavtype == "TravellingWave":
                for param in tw_required_attrs:
                    if param not in (h5file.attrs.keys()):
                        raise KeyError(
                            f"{param} must be an attribute of {filename} for a travelling wave linac"
                        )
                    else:
                        setattr(self, param, h5file.attrs[param])
        length_set = (
            False if self.field_type != "2DElectroDynamic" else h5file.attrs["length"]
        )
        for key in h5file:
            if not length_set:
                self.length = len(h5file[key][()])
            setattr(
                self,
                key,
                FieldParameter(
                    name=key,
                    value=UnitValue(h5file[key][()], units=h5file[key].attrs["units"]),
                ),
            )
        # if ("z" not in list(h5file.keys())) and ("t" in list(h5file.keys())):
        #     setattr(self, "z", self.t.value * speed_of_light)
        for param in ["Ex", "Ey", "Ez", "Er", "Bx", "By", "Bz", "Br", "G"]:
            if getattr(self, param).value is not None:
                lessthancond = 0.99 > round(max(abs(getattr(self, param).value.val)), 4)
                greaterthancond = (
                    round(max(abs(getattr(self, param).value.val)), 4) > 1.1
                )
                if lessthancond or greaterthancond:
                    warn(
                        f"Warning: Field {param} in {filename} is not normalised to 1.0."
                    )
        self.read = True
    return filename


def write_HDF5_field_file(self):
    """
    Write the :class:`nala.translator.utils.fields.field` object to an HDF5 file.
    All of the attributes of the class are read, and if they are defined and in the correct format,
    they are written to the file.

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object to be saved.

    Returns
    -------
    str:
        The name of the HDF5 file
    """
    self.filename = os.path.splitext(self.filename)[0] + ".hdf5"
    with h5py.File(self.filename, "w") as h5file:
        attrs = [
            "field_type",
            "frequency",
            "radius",
            "fourier",
            "cavity_type",
            "start_cell_z",
            "end_cell_z",
            "mode_numerator",
            "mode_denominator",
            "orientation",
            "length",
        ]
        for att in attrs:
            if hasattr(self, att):
                if getattr(self, att) is not None:
                    h5file.attrs[att] = getattr(self, att)
        dsets = [
            "x",
            "y",
            "z",
            "r",
            "Ex",
            "Ey",
            "Ez",
            "Er",
            "Bx",
            "By",
            "Bz",
            "Br",
            "Wx",
            "Wy",
            "Wz",
            "Wr",
            "G",
        ]
        for dset in dsets:
            if isinstance(getattr(self, dset), FieldParameter):
                if getattr(self, dset).value is not None:
                    dataset = h5file.create_dataset(
                        dset, data=getattr(self, dset).value
                    )
                    dataset.attrs["units"] = getattr(self, dset).value.units
    return self.filename



import unittest
from nala.models.element import Element, ElectricalElement, ManufacturerElement


class TestElement(unittest.TestCase):

    def test_element_creation_with_no_controls(self):
        elem = Element(
            machine_area="AREA-01",
            hardware_type="Magnet",
            hardware_class="Quadrupole",
            name="CLA-A1-QUAD-01",
        )
        self.assertEqual(elem.machine_area, "AREA-01")
        self.assertEqual(elem.hardware_type, "Magnet")
        self.assertEqual(elem.name, "CLA-A1-QUAD-01")
        self.assertIsInstance(elem, Element)
        self.assertIsInstance(elem.electrical, ElectricalElement)
        self.assertIsInstance(elem.manufacturer, ManufacturerElement)
        self.assertIsNone(elem.controls, None)
        self.assertIsNone(elem.alias)

    def test_element_with_aliases(self):
        elem = Element(
            machine_area="AREA-01",
            hardware_type="Magnet",
            hardware_class="Quadrupole",
            name="CLA-A1-QUAD-01",
            alias=["CLA-QUAD01", "CLA-QUAD1"],
        )
        self.assertEqual(elem.alias.aliases, ["CLA-QUAD01", "CLA-QUAD1"])

    def test_to_CATAP(self):
        elem = Element(
            machine_area="AREA-01",
            hardware_type="Magnet",
            hardware_class="Quadrupole",
            name="CLA-A1-QUAD-01",
            alias=["CLA-QUAD01", "CLA-QUAD1"],
        )
        catap_dict = elem.to_CATAP()
        self.assertEqual(catap_dict["machine_area"], "AREA-01")
        self.assertEqual(catap_dict["hardware_type"], "Magnet")
        self.assertEqual(catap_dict["name"], "CLA-A1-QUAD-01")
        self.assertEqual(catap_dict["name_alias"], ["CLA-QUAD01", "CLA-QUAD1"])
        self.assertEqual(catap_dict["manufacturer"], "")
        self.assertNotIn("controls", catap_dict)
        self.assertNotIn("electrical", catap_dict)

    def test_element_creation_with_controls(self):
        controls_info = {
            "variables": {
                "var1": {
                    "identifier": "var1",
                    "dtype": "float",
                    "protocol": "CA",
                    "units": "V",
                    "description": "A float variable for voltage",
                }
            }
        }
        elem = Element(
            machine_area="AREA-01",
            hardware_type="Magnet",
            hardware_class="Quadrupole",
            name="CLA-A1-QUAD-01",
            controls=controls_info,
        )
        self.assertIsNotNone(elem.controls)
        self.assertIn("var1", elem.controls.variables)
        self.assertEqual(elem.controls.variables["var1"].identifier, "var1")
        self.assertEqual(elem.controls.variables["var1"].dtype, float)
        self.assertEqual(elem.controls.variables["var1"].protocol, "CA")
        self.assertEqual(elem.controls.variables["var1"].units, "V")
        self.assertEqual(
            elem.controls.variables["var1"].description, "A float variable for voltage"
        )

    def test_element_serialisation(self):
        controls_info = {
            "variables": {
                "var1": {
                    "identifier": "var1",
                    "dtype": "float",
                    "protocol": "CA",
                    "units": "V",
                    "description": "A float variable for voltage",
                }
            }
        }
        elem = Element(
            machine_area="AREA-01",
            hardware_type="Magnet",
            hardware_class="Quadrupole",
            name="CLA-A1-QUAD-01",
            controls=controls_info,
        )
        serialized = elem.model_dump()
        self.assertIn("machine_area", serialized)
        self.assertIn("hardware_type", serialized)
        self.assertIn("hardware_class", serialized)
        self.assertIn("name", serialized)
        self.assertIn("manufacturer", serialized)
        self.assertIn("electrical", serialized)
        self.assertIn("alias", serialized)
        self.assertIn("physical", serialized)
        self.assertIn("controls", serialized)
        self.assertIn("variables", serialized["controls"])
        self.assertIn("var1", serialized["controls"]["variables"])
        self.assertEqual(
            serialized["controls"]["variables"]["var1"]["identifier"], "var1"
        )
        self.assertEqual(serialized["controls"]["variables"]["var1"]["dtype"], "float")
        self.assertEqual(serialized["controls"]["variables"]["var1"]["protocol"], "CA")
        self.assertEqual(serialized["controls"]["variables"]["var1"]["units"], "V")
        self.assertEqual(
            serialized["controls"]["variables"]["var1"]["description"],
            "A float variable for voltage",
        )


if __name__ == "__main__":
    unittest.main()



from cheetah.accelerator import (  # noqa: F401
    BPM,
    Aperture,
    Cavity,
    CustomTransferMap,
    Dipole,
    Drift,
    Element,
    HorizontalCorrector,
    Marker,
    Quadrupole,
    RBend,
    Screen,
    Segment,
    Sextupole,
    Solenoid,
    SpaceChargeKick,
    TransverseDeflectingCavity,
    Undulator,
    VerticalCorrector,
)

cheetah_conversion_rules = {
    "Dipole": Dipole,
    "Solenoid": Solenoid,
    "Quadrupole": Quadrupole,
    "Sextupole": Sextupole,
    "Octupole": Drift,
    "Beam_Position_Monitor": Screen,
    "Beam_Arrival_Monitor": Screen,
    "Bunch_Length_Monitor": Screen,
    "Screen": Screen,
    "RCollimator": Aperture,
    "Collimator": Aperture,
    "Monitor": Screen,
    "Wall_Current_Monitor": Drift,
    "Integrated_Current_Transformer": Drift,
    "Faraday_Cup_Monitor": Drift,
    "RFCavity": Cavity,
    "RFDeflectingCavity": TransverseDeflectingCavity,
    "Aperture": Aperture,
    "Shutter": Drift,
    "Valve": Drift,
    "Bellows": Drift,
    "Cleaner": Drift,
    "lscdrift": Drift,
    "Drift": Drift,
    "csrdrift": Drift,
    "Kicker": HorizontalCorrector,
    "Combined_Corrector": HorizontalCorrector,
    "Marker": Screen,
    "Wakefield": Drift,
    "Laser": Drift,
}



from .base import BaseElementTranslator
from nala.models.simulation import TwissMatchSimulationElement

class TwissMatchTranslator(BaseElementTranslator):
    """
    Translator class for converting a :class:`~nala.models.element.TwissMatch` element instance into a string or
    object that can be understood by various simulation codes.
    """

    simulation: TwissMatchSimulationElement
    """Twiss match simulation element"""


import sys
import yaml
from pysdds import read as sddsread

sys.path.append("..")
from PAdantic.PAdantic import PAdantic  # noqa E402

machine = PAdantic(
    layout_file="Machines/CLARA/layouts.yaml",
    section_file="Machines/CLARA/sections.yaml",
    yaml_dir="Machines/CLARA/YAML/",
)

machine.default_layout = "CLARA"

simframe_examples = "../../SimFrame_Examples/FEBE/StandardSetups/Setups/Setup_1_250pC"

with open(
    f"{simframe_examples}/lattice.yaml",
    "r",
) as stream:
    try:
        lattice = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)

sdds_twiss = {}
sdds_twiss_file = sddsread(f"{simframe_examples}/FEBE.twi")
for idx, name in enumerate(sdds_twiss_file.col("ElementName").data[0]):
    sdds_twiss[name] = 0.511 * sdds_twiss_file.col("pCentral0").data[0][idx]

current_dictionary = {}
for name, elem in lattice["elements"].items():
    if (
        elem["type"] == "quadrupole"
        and "-FEA-" not in name
        and "-FEH-" not in name
        and "-FED-" not in name
    ):
        linsat = machine.get_element(name).magnetic.linear_saturation_coefficients
        momentum = sdds_twiss[name]
        k1 = elem["k1l"] / elem["length"]
        print(name, k1, momentum, linsat.KToCurrent(k1, momentum))
        current_dictionary[name] = float(linsat.KToCurrent(k1, momentum))

with open("test_currents.yaml", "w") as file:
    yaml.dump(current_dictionary, file)



import numpy as np
from scipy.constants import speed_of_light, pi
from pydantic import (
    BaseModel,
    model_serializer,
    Field,
    field_validator,
    NonNegativeInt,
    create_model,
    NonNegativeFloat,
    computed_field,
)
from .baseModels import IgnoreExtra, T

class PlasmaElement(IgnoreExtra):
    """Plasma model."""

    density: float = Field(gt=0)
    """Plasma density in m^-3"""

    species: str = "electron"
    """Plasma species (e.g., 'electron', 'proton', etc.)
    # TODO make these literals when we have a list of supported species"""

    ramp_up: float = Field(ge=0, default=0.001)
    """Plasma ramp length at entrance [metres]."""

    plateau: float = Field(ge=0, default=0.001)
    """Plasma plateau length [metres]."""

    ramp_down: float = Field(ge=0, default=0.001)
    """Plasma ramp length at exit [metres]."""

    ramp_decay_length: float = Field(ge=0, default=0.001)
    """Plasma decay length [metres]."""

    density_profile: bool = False
    """Density profile function; if False, a flat profile is used; if True, use
     the :func:`~_density_profile` method to calculate the density profile based on
     `ramp_up`, `plateau`, `ramp_down` and `ramp_decay_length`. Only linear profiles """

    parabolic_coefficient: float = 0
    """Parabolic coefficient for density profile"""